#    Copyright 2013 IBM Corp.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Common internal object model"""

from __future__ import annotations

import abc
import collections
from collections.abc import (
    Callable,
    Iterator,
    MutableMapping,
    Mapping,
    Sequence,
)
import copy
import functools
import logging
from typing import (
    Any,
    cast,
    Concatenate,
    overload,
    ParamSpec,
    Self,
    TypeVar,
)
import warnings

import oslo_messaging as messaging
from oslo_utils import excutils
from oslo_utils import versionutils as vutils

from oslo_versionedobjects._i18n import _
from oslo_versionedobjects import exception
from oslo_versionedobjects import fields as obj_fields

LOG = logging.getLogger('object')

_VO = TypeVar('_VO', bound='VersionedObject')
# we'd like this to be ObjectListBase[_VO] but Python doesn't allow this :(
_VOL = TypeVar('_VOL', bound='ObjectListBase[Any]')


class _NotSpecifiedSentinel:
    pass


def _get_attrname(name: str) -> str:
    """Return the mangled name of the attribute's underlying storage."""
    return '_obj_' + name


def _make_class_properties(cls: type[VersionedObject]) -> None:
    # NOTE(danms/comstud): Inherit fields from super classes.
    # mro() returns the current class first and returns 'object' last, so
    # those can be skipped.  Also be careful to not overwrite any fields
    # that already exist.  And make sure each cls has its own copy of
    # fields and that it is not sharing the dict with a super class.
    cls.fields = dict(cls.fields)
    for supercls in cls.mro()[1:-1]:
        if not hasattr(supercls, 'fields'):
            continue
        for name, field in supercls.fields.items():
            if name not in cls.fields:
                cls.fields[name] = field
    for name, field in cls.fields.items():
        if not isinstance(field, obj_fields.Field):
            raise exception.ObjectFieldInvalid(
                field=name, objname=cls.obj_name()
            )

        def getter(self: VersionedObject, name: str = name) -> Any:
            attrname = _get_attrname(name)
            if not hasattr(self, attrname):
                self.obj_load_attr(name)
            return getattr(self, attrname)

        def setter(
            self: VersionedObject,
            value: Any,
            name: str = name,
            field: obj_fields.Field[Any] = field,
        ) -> None:
            attrname = _get_attrname(name)
            field_value = field.coerce(self, name, value)
            if field.read_only and hasattr(self, attrname):
                # Note(yjiang5): _from_db_object() may iterate
                # every field and write, no exception in such situation.
                if getattr(self, attrname) != field_value:
                    raise exception.ReadOnlyFieldError(field=name)
                else:
                    return

            self._changed_fields.add(name)
            try:
                setattr(self, attrname, field_value)
            except Exception:
                with excutils.save_and_reraise_exception():
                    attr = f"{self.obj_name()}.{name}"
                    LOG.exception('Error setting %(attr)s', {'attr': attr})

        def deleter(self: VersionedObject, name: str = name) -> None:
            attrname = _get_attrname(name)
            if not hasattr(self, attrname):
                raise AttributeError(f"No such attribute `{name}'")
            delattr(self, attrname)

        setattr(cls, name, property(getter, setter, deleter))


class VersionedObjectRegistry:
    _registry: VersionedObjectRegistry | None = None
    _obj_classes: collections.defaultdict[str, list[type[VersionedObject]]]

    def __new__(cls, *args: Any, **kwargs: Any) -> Self:
        if not VersionedObjectRegistry._registry:
            VersionedObjectRegistry._registry = object.__new__(
                VersionedObjectRegistry, *args, **kwargs
            )
            VersionedObjectRegistry._registry._obj_classes = (
                collections.defaultdict(list)
            )
        self = object.__new__(cls, *args, **kwargs)
        self._obj_classes = VersionedObjectRegistry._registry._obj_classes
        return self

    def registration_hook(
        self, cls: type[VersionedObject], index: int
    ) -> None:
        pass

    def _register_class(self, cls: type[VersionedObject]) -> None:
        def _vers_tuple(
            obj: type[VersionedObject],
        ) -> tuple[int, ...]:
            return vutils.convert_version_to_tuple(obj.VERSION)

        _make_class_properties(cls)
        obj_name = cls.obj_name()
        for i, obj in enumerate(self._obj_classes[obj_name]):
            self.registration_hook(cls, i)
            if cls.VERSION == obj.VERSION:
                self._obj_classes[obj_name][i] = cls
                break
            if _vers_tuple(cls) > _vers_tuple(obj):
                # Insert before.
                self._obj_classes[obj_name].insert(i, cls)
                break
        else:
            # Either this is the first time we've seen the object or it's
            # an older version than anything we'e seen.
            self._obj_classes[obj_name].append(cls)
            self.registration_hook(cls, 0)

    @classmethod
    def register(cls, obj_cls: type[_VO]) -> type[_VO]:
        registry = cls()
        registry._register_class(obj_cls)
        return obj_cls

    @classmethod
    def register_if(cls, condition: bool) -> Callable[[type[_VO]], type[_VO]]:
        def wraps(obj_cls: type[_VO]) -> type[_VO]:
            if condition:
                obj_cls = cls.register(obj_cls)
            else:
                _make_class_properties(obj_cls)
            return obj_cls

        return wraps

    @classmethod
    def objectify(cls, obj_cls: type[_VO]) -> type[_VO]:
        return cls.register_if(False)(obj_cls)

    @classmethod
    def obj_classes(
        cls,
    ) -> collections.defaultdict[str, list[type[VersionedObject]]]:
        registry = cls()
        return registry._obj_classes


P = ParamSpec('P')
R = TypeVar('R')


# Decorator that marks an object's method as remotable.
# If the metaclass is configured to forward object methods to an
# indirection service, this will result in making an RPC call
# instead of directly calling the implementation in the object. Instead,
# the object implementation on the remote end will perform the
# requested action and the result will be returned here.
#
# This decorator works for both instance methods and classmethods. When
# used with an instance method, context is retrieved from self._context:
#
#     @remotable
#     def foo(self):
#         ...
#
# When used with a classmethod, context is passed as first argument after cls:
#
#       @classmethod
#       @remotable
#       def bar(cls, context):
#           ...
#
# Note that for instance methods, if context is not present, the object
# is "orphaned" and remotable methods cannot be called.
@overload
def remotable(
    fn: Callable[Concatenate[_VO, P], R],
) -> Callable[Concatenate[_VO, P], R]: ...


@overload
def remotable(
    fn: Callable[Concatenate[type[_VO], P], R],
) -> Callable[Concatenate[type[_VO], P], R]: ...


def remotable(fn: Callable[..., R]) -> Callable[..., R]:
    """Decorator for remotable object methods."""

    def _wrapper_classmethod(cls: type[_VO], *args: Any, **kwargs: Any) -> R:
        if args:
            context = args[0]
            remaining_args = args[1:]
            remaining_kwargs: dict[str, object] = kwargs
        else:
            remaining_args = ()
            remaining_kwargs = kwargs.copy()
            context = remaining_kwargs.pop('context')
        if cls.indirection_api:
            version_manifest = obj_tree_get_versions(cls.obj_name())
            try:
                result = cls.indirection_api.object_class_action_versions(
                    context,
                    cls.obj_name(),
                    fn.__name__,
                    version_manifest,
                    remaining_args,
                    remaining_kwargs,
                )
            except NotImplementedError:
                result = cls.indirection_api.object_class_action(
                    context,
                    cls.obj_name(),
                    fn.__name__,
                    cls.VERSION,
                    remaining_args,
                    remaining_kwargs,
                )
            return cast(R, result)

        result = fn(cls, *args, **kwargs)
        if isinstance(result, VersionedObject):
            result._context = context
        return result

    def _wrapper_instancemethod(self: _VO, *args: Any, **kwargs: Any) -> R:
        ctxt = self._context
        if ctxt is None:
            raise exception.OrphanedObjectError(
                method=fn.__name__, objtype=self.obj_name()
            )

        if self.indirection_api:
            updates, result = self.indirection_api.object_action(
                ctxt, self, fn.__name__, args, kwargs
            )
            for key, value in updates.items():
                if key in self.fields:
                    field = self.fields[key]
                    # NOTE(ndipanov): Since VersionedObjectSerializer will
                    # have deserialized any object fields into objects
                    # already, we do not try to deserialize them again.
                    if isinstance(value, VersionedObject):
                        setattr(self, key, value)
                    else:
                        setattr(
                            self,
                            key,
                            field.from_primitive(self, key, value),
                        )
            self.obj_reset_changes()
            self._changed_fields = set(updates.get('obj_what_changed', []))
            return cast(R, result)

        return fn(self, *args, **kwargs)

    def wrapper(
        first: type[_VO] | _VO,
        /,
        *args: Any,
        **kwargs: Any,
    ) -> R:
        if isinstance(first, type):
            return _wrapper_classmethod(first, *args, **kwargs)
        else:
            return _wrapper_instancemethod(first, *args, **kwargs)

    # NOTE(danms): Make this discoverable
    setattr(wrapper, 'remotable', True)
    setattr(wrapper, 'original_fn', fn)

    functools.update_wrapper(wrapper, fn)

    return wrapper


def remotable_classmethod(
    fn: Callable[Concatenate[type[_VO], P], R],
) -> classmethod[_VO, P, R]:
    """Decorator for remotable classmethods.

    .. deprecated::
        Use ``@classmethod`` combined with ``@remotable`` instead::

            @classmethod
            @remotable
            def my_method(cls, context): ...
    """
    warnings.warn(
        "remotable_classmethod is deprecated, use @classmethod with "
        "@remotable",
        DeprecationWarning,
        stacklevel=2,
    )

    def wrapper(cls: type[_VO], /, *args: P.args, **kwargs: P.kwargs) -> R:
        if args:
            context = args[0]
            remaining_args = args[1:]
            remaining_kwargs: dict[str, object] = kwargs
        else:
            remaining_args = ()
            remaining_kwargs = kwargs.copy()
            context = remaining_kwargs.pop('context')
        if cls.indirection_api:
            version_manifest = obj_tree_get_versions(cls.obj_name())
            try:
                result = cls.indirection_api.object_class_action_versions(
                    context,
                    cls.obj_name(),
                    fn.__name__,
                    version_manifest,
                    remaining_args,
                    remaining_kwargs,
                )
            except NotImplementedError:
                # FIXME(danms): Maybe start to warn here about deprecation?
                result = cls.indirection_api.object_class_action(
                    context,
                    cls.obj_name(),
                    fn.__name__,
                    cls.VERSION,
                    remaining_args,
                    remaining_kwargs,
                )
            return cast(R, result)

        result = fn(cls, *args, **kwargs)
        if isinstance(result, VersionedObject):
            result._context = context
        return result

    # NOTE(danms): Make this discoverable
    setattr(wrapper, 'remotable', True)
    setattr(wrapper, 'original_fn', fn)

    functools.update_wrapper(wrapper, fn)

    return classmethod(wrapper)


class VersionedObject:
    """Base class and object factory.

    This forms the base of all objects that can be remoted or instantiated
    via RPC. Simply defining a class that inherits from this base class
    will make it remotely instantiatable. Objects should implement the
    necessary "get" classmethod routines as well as "save" object methods
    as appropriate.
    """

    indirection_api: VersionedObjectIndirectionAPI | None = None

    #: Object version field
    #:
    #: Each service has its set of objects, each with a version attached. When
    #: a client attempts to call an object method, the server checks to see if
    #: the version of that object matches (in a compatible way) its object
    #: implementation. If so, cool, and if not, fail.
    #:
    #: This version is allowed to have three parts, ``X.Y.Z``, where the ``.Z``
    #: element is reserved for stable branch backports. The ``.Z`` is ignored
    #: for the purposes of triggering a backport, which means anything changed
    #: under a ``.Z`` must be additive and non-destructive such that a node
    #: that knows about ``X.Y`` can consider ``X.Y.Z`` equivalent.
    VERSION: str = '1.0'

    #: Object namespace for serialization
    #:
    #: This should generally not be changed, but is needed for backwards
    #: compatibility
    OBJ_SERIAL_NAMESPACE: str = 'versioned_object'

    #: Object project namespace for serialization
    #:
    #: This is used to disambiguate owners of objects sharing a common RPC
    #: medium.
    OBJ_PROJECT_NAMESPACE: str = 'versionedobjects'

    #: The fields present in this object as ``key:field`` pairs.
    #:
    #: For example::
    #:
    #:     fields = {
    #:         'foo': obj_fields.IntegerField(),
    #:         'bar': obj_fields.StringField(),
    #:     }
    #:
    fields: MutableMapping[str, obj_fields.Field[Any]] = {}

    #: List of additional object attributes to expose alongside :attr:`fields`.
    #:
    #: Names listed here are typically implemented as Python properties and
    #: are included in :attr:`obj_fields` and exposed through iteration, but
    #: they are not versioned and not serialized during RPC::
    #:
    #:     class MyObject(VersionedObject):
    #:         fields = {'foo': obj_fields.IntegerField()}
    #:         obj_extra_fields = ['bar']
    #:
    #:         @property
    #:         def bar(self):
    #:             return self.foo * 2
    obj_extra_fields: Sequence[str] = []

    #: Table of sub-object versioning information
    #:
    #: This contains a list of version mappings, by the field name of
    #: the subobject. The mappings must be in order of oldest to
    #: newest, and are tuples of ``(my_version, subobject_version)``. A
    #: request to backport this object to ``my_version`` will cause the
    #: subobject to be backported to ``subobject_version``::
    #:
    #:     obj_relationships = {
    #:         'subobject1': [('1.2', '1.1'), ('1.4', '1.2')],
    #:         'subobject2': [('1.2', '1.0')],
    #:     }
    #:
    #: In the above example:
    #:
    #: * If we are asked to backport our object to version 1.3,
    #:   ``subobject1`` will be backported to version 1.1, since it was
    #:   bumped to version 1.2 when our version was 1.4.
    #: * If we are asked to backport our object to version 1.5,
    #:   no changes will be made to ``subobject1`` or ``subobject2``, since
    #:   they have not changed since version 1.4.
    #: * If we are asked to backlevel our object to version 1.1, we
    #:   will remove both ``subobject1`` and ``subobject2`` from the primitive,
    #:   since they were not added until version 1.2.
    #:
    obj_relationships: dict[str, list[tuple[str, str]]] = {}

    _changed_fields: set[str]
    _context: Any

    def __init__(self, context: Any = None, **kwargs: Any) -> None:
        self._changed_fields = set()
        self._context = context
        for key in kwargs.keys():
            setattr(self, key, kwargs[key])

    def __repr__(self) -> str:
        repr_str = '{}({})'.format(
            self.obj_name(),
            ','.join(
                [
                    '{}={}'.format(
                        name,
                        (
                            self.obj_attr_is_set(name)
                            and field.stringify(getattr(self, name))
                            or '<?>'
                        ),
                    )
                    for name, field in sorted(self.fields.items())
                ]
            ),
        )
        return repr_str

    def __contains__(self, name: object) -> bool:
        if not isinstance(name, str):
            return False
        try:
            return self.obj_attr_is_set(name)
        except AttributeError:
            return False

    @classmethod
    def to_json_schema(cls) -> dict[str, Any]:
        obj_name = cls.obj_name()
        schema: dict[str, Any] = {
            '$schema': 'http://json-schema.org/draft-04/schema#',
            'title': obj_name,
        }
        schema.update(obj_fields.Object(obj_name).get_schema())
        return schema

    @classmethod
    def obj_name(cls) -> str:
        """Return the object's name.

        Return a canonical name for this object which will be used over
        the wire for remote hydration.
        """
        return cls.__name__

    @classmethod
    def _obj_primitive_key(cls, field: str) -> str:
        return f'{cls.OBJ_SERIAL_NAMESPACE}.{field}'

    @classmethod
    def _obj_primitive_field(
        cls,
        primitive: dict[str, Any],
        field: str,
        default: Any = obj_fields.UnspecifiedDefault,
    ) -> Any:
        key = cls._obj_primitive_key(field)
        if default == obj_fields.UnspecifiedDefault:
            return primitive[key]
        else:
            return primitive.get(key, default)

    @classmethod
    def obj_class_from_name(
        cls, objname: str, objver: str
    ) -> type[VersionedObject]:
        """Returns a class from the registry based on a name and version."""
        if objname not in VersionedObjectRegistry.obj_classes():
            LOG.error(
                'Unable to instantiate unregistered object type %(objtype)s',
                {'objtype': objname},
            )
            raise exception.UnsupportedObjectError(objtype=objname)

        # NOTE(comstud): If there's not an exact match, return the highest
        # compatible version. The objects stored in the class are sorted
        # such that highest version is first, so only set compatible_match
        # once below.
        compatible_match: type[VersionedObject] | None = None

        for objclass in VersionedObjectRegistry.obj_classes()[objname]:
            if objclass.VERSION == objver:
                return objclass
            if not compatible_match and vutils.is_compatible(
                objver, objclass.VERSION
            ):
                compatible_match = objclass

        if compatible_match:
            return compatible_match

        # As mentioned above, latest version is always first in the list.
        latest_ver = VersionedObjectRegistry.obj_classes()[objname][0].VERSION
        raise exception.IncompatibleObjectVersion(
            objname=objname, objver=objver, supported=latest_ver
        )

    @classmethod
    def _obj_from_primitive(
        cls, context: Any, objver: str, primitive: dict[str, Any]
    ) -> Self:
        self = cls()
        self._context = context
        self.VERSION = objver
        objdata = cls._obj_primitive_field(primitive, 'data')
        changes = cls._obj_primitive_field(primitive, 'changes', [])
        for name, field in self.fields.items():
            if name in objdata:
                setattr(
                    self, name, field.from_primitive(self, name, objdata[name])
                )
        self._changed_fields = {x for x in changes if x in self.fields}
        return self

    @classmethod
    def obj_from_primitive(
        cls, primitive: dict[str, Any], context: Any = None
    ) -> Self:
        """Object field-by-field hydration."""
        objns = cls._obj_primitive_field(primitive, 'namespace')
        objname = cls._obj_primitive_field(primitive, 'name')
        objver = cls._obj_primitive_field(primitive, 'version')
        if objns != cls.OBJ_PROJECT_NAMESPACE:
            # NOTE(danms): We don't do anything with this now, but it's
            # there for "the future"
            raise exception.UnsupportedObjectError(
                objtype=f'{objns}.{objname}'
            )
        objclass = cls.obj_class_from_name(objname, objver)
        return objclass._obj_from_primitive(context, objver, primitive)  # type: ignore[return-value]

    def __deepcopy__(self, memo: dict[int, Any]) -> Self:
        """Efficiently make a deep copy of this object."""

        # NOTE(danms): A naive deepcopy would copy more than we need,
        # and since we have knowledge of the volatile bits of the
        # object, we can be smarter here. Also, nested entities within
        # some objects may be uncopyable, so we can avoid those sorts
        # of issues by copying only our field data.

        nobj = self.__class__()

        # NOTE(sskripnick): we should save newly created object into mem
        # to let deepcopy know which branches are already created.
        # See launchpad bug #1602314 for more details
        memo[id(self)] = nobj
        nobj._context = self._context
        for name in self.fields:
            if self.obj_attr_is_set(name):
                nval = copy.deepcopy(getattr(self, name), memo)
                setattr(nobj, name, nval)
        nobj._changed_fields = set(self._changed_fields)
        return nobj

    def obj_clone(self) -> Self:
        """Create a copy."""
        return copy.deepcopy(self)

    def _obj_relationship_for(
        self, field: str, target_version: str
    ) -> list[tuple[str, str]] | None:
        # NOTE(danms): We need to be graceful about not having the temporary
        # version manifest if called from obj_make_compatible().
        if (
            not hasattr(self, '_obj_version_manifest')
            or self._obj_version_manifest is None
        ):
            try:
                return self.obj_relationships[field]
            except KeyError:
                raise exception.ObjectActionError(
                    action='obj_make_compatible', reason=f'No rule for {field}'
                )

        # The field is expected to be an ObjectField or ListOfObjectsField;
        # both store objname as an instance attribute set in __init__
        objname = cast(obj_fields.ObjectField, self.fields[field]).objname
        if objname not in self._obj_version_manifest:
            return None
        # NOTE(danms): Compute a relationship mapping that looks like
        # what the caller expects.
        return [(target_version, self._obj_version_manifest[objname])]

    def _obj_make_obj_compatible(
        self, primitive: dict[str, Any], target_version: str, field: str
    ) -> None:
        """Backlevel a sub-object based on our versioning rules.

        This is responsible for backporting objects contained within
        this object's primitive according to a set of rules we
        maintain about version dependencies between objects. This
        requires that the obj_relationships table in this object is
        correct and up-to-date.

        :param primitive: The primitive version of this object
        :param target_version: The version string requested for this object
        :param field: The name of the field in this object containing the
                      sub-object to be backported
        """
        relationship_map = self._obj_relationship_for(field, target_version)
        if not relationship_map:
            # NOTE(danms): This means the field was not specified in the
            # version manifest from the client, so it must not want this
            # field, so skip.
            return

        try:
            _get_subobject_version(
                target_version,
                relationship_map,
                lambda ver: _do_subobject_backport(
                    ver, self, field, primitive
                ),
            )
        except exception.TargetBeforeSubobjectExistedException:
            # Subobject did not exist, so delete it from the primitive
            del primitive[field]

    def obj_make_compatible(
        self, primitive: dict[str, Any], target_version: str
    ) -> None:
        """Make an object representation compatible with a target version.

        This is responsible for taking the primitive representation of
        an object and making it suitable for the given target_version.
        This may mean converting the format of object attributes, removing
        attributes that have been added since the target version, etc. In
        general:

        - If a new version of an object adds a field, this routine
          should remove it for older versions.
        - If a new version changed or restricted the format of a field, this
          should convert it back to something a client knowing only of the
          older version will tolerate.
        - If an object that this object depends on is bumped, then this
          object should also take a version bump. Then, this routine should
          backlevel the dependent object (by calling its obj_make_compatible())
          if the requested version of this object is older than the version
          where the new dependent object was added.

        :param primitive: The result of :meth:`obj_to_primitive`
        :param target_version: The version string requested by the recipient
                               of the object
        :raises: :exc:`oslo_versionedobjects.exception.UnsupportedObjectError`
                 if conversion is not possible for some reason
        """
        for key, field in self.fields.items():
            if not isinstance(
                field, (obj_fields.ObjectField, obj_fields.ListOfObjectsField)
            ):
                continue
            if not self.obj_attr_is_set(key):
                continue
            self._obj_make_obj_compatible(primitive, target_version, key)

    def obj_make_compatible_from_manifest(
        self,
        primitive: dict[str, Any],
        target_version: str,
        version_manifest: dict[str, str] | None,
    ) -> None:
        # NOTE(danms): Stash the manifest on the object so we can use it in
        # the deeper layers. We do this because obj_make_compatible() is
        # defined library API at this point, yet we need to get this manifest
        # to the other bits that get called so we can propagate it to child
        # calls. It's not pretty, but a tactical solution. Ideally we will
        # either evolve or deprecate obj_make_compatible() in a major version
        # bump.
        self._obj_version_manifest = version_manifest
        try:
            return self.obj_make_compatible(primitive, target_version)
        finally:
            delattr(self, '_obj_version_manifest')

    def obj_to_primitive(
        self,
        target_version: str | None = None,
        version_manifest: dict[str, str] | None = None,
    ) -> dict[str, Any]:
        """Simple base-case dehydration.

        This calls to_primitive() for each item in fields.
        """
        if target_version is None:
            target_version = self.VERSION
        if vutils.convert_version_to_tuple(
            target_version
        ) > vutils.convert_version_to_tuple(self.VERSION):
            raise exception.InvalidTargetVersion(version=target_version)
        primitive = dict()
        for name, field in self.fields.items():
            if self.obj_attr_is_set(name):
                primitive[name] = field.to_primitive(
                    self, name, getattr(self, name)
                )
        # NOTE(danms): If we know we're being asked for a different version,
        # then do the compat step. However, even if we think we're not,
        # we may have sub-objects that need it, so if we have a manifest we
        # have to traverse this object just in case. Previously, we
        # required a parent version bump for any child, so the target
        # check was enough.
        if target_version != self.VERSION or version_manifest:
            self.obj_make_compatible_from_manifest(
                primitive, target_version, version_manifest
            )
        obj = {
            self._obj_primitive_key('name'): self.obj_name(),
            self._obj_primitive_key('namespace'): (self.OBJ_PROJECT_NAMESPACE),
            self._obj_primitive_key('version'): target_version,
            self._obj_primitive_key('data'): primitive,
        }
        if self.obj_what_changed():
            # NOTE(cfriesen): if we're downgrading to a lower version, then
            # it's possible that self.obj_what_changed() includes fields that
            # no longer exist in the lower version.  If so, filter them out.
            what_changed = self.obj_what_changed()
            changes = [field for field in what_changed if field in primitive]
            if changes:
                obj[self._obj_primitive_key('changes')] = changes
        return obj

    def obj_set_defaults(self, *attrs: str) -> None:
        if not attrs:
            attrs = tuple(
                name
                for name, field in self.fields.items()
                if field.default != obj_fields.UnspecifiedDefault
            )

        for attr in attrs:
            default = copy.deepcopy(self.fields[attr].default)
            if default is obj_fields.UnspecifiedDefault:
                raise exception.ObjectActionError(
                    action='set_defaults',
                    reason=f'No default set for field {attr}',
                )
            if not self.obj_attr_is_set(attr):
                setattr(self, attr, default)

    def obj_load_attr(self, attrname: str) -> None:
        """Load an additional attribute from the real object.

        This should load self.$attrname and cache any data that might
        be useful for future load operations.
        """
        raise NotImplementedError(
            _("Cannot load '%s' in the base class") % attrname
        )

    def save(self) -> None:
        """Save the changed fields back to the store.

        This is optional for subclasses, but is presented here in the base
        class for consistency among those that do.
        """
        raise NotImplementedError(_('Cannot save anything in the base class'))

    def obj_what_changed(self) -> set[str]:
        """Returns a set of fields that have been modified."""
        changes = {
            field for field in self._changed_fields if field in self.fields
        }
        for field in self.fields:
            if (
                self.obj_attr_is_set(field)
                and isinstance(getattr(self, field), VersionedObject)
                and getattr(self, field).obj_what_changed()
            ):
                changes.add(field)
        return changes

    def obj_get_changes(self) -> dict[str, Any]:
        """Returns a dict of changed fields and their new values."""
        changes: dict[str, Any] = {}
        for key in self.obj_what_changed():
            changes[key] = getattr(self, key)
        return changes

    def obj_reset_changes(
        self, fields: list[str] | None = None, recursive: bool = False
    ) -> None:
        """Reset the list of fields that have been changed.

        :param fields: List of fields to reset, or "all" if None.
        :param recursive: Call obj_reset_changes(recursive=True) on
                          any sub-objects within the list of fields
                          being reset.

        This is NOT "revert to previous values".

        Specifying fields on recursive resets will only be honored at the top
        level. Everything below the top will reset all.
        """
        if recursive:
            for field in self.obj_get_changes():
                # Ignore fields not in requested set (if applicable)
                if fields and field not in fields:
                    continue

                # Skip any fields that are unset
                if not self.obj_attr_is_set(field):
                    continue

                value = getattr(self, field)

                # Don't reset nulled fields
                if value is None:
                    continue

                # Reset straight Object and ListOfObjects fields
                if isinstance(self.fields[field], obj_fields.ObjectField):
                    value.obj_reset_changes(recursive=True)
                elif isinstance(
                    self.fields[field], obj_fields.ListOfObjectsField
                ):
                    for thing in value:
                        thing.obj_reset_changes(recursive=True)

        if fields:
            self._changed_fields -= set(fields)
        else:
            self._changed_fields.clear()

    def obj_attr_is_set(self, attrname: str) -> bool:
        """Test object to see if attrname is present.

        Returns True if the named attribute has a value set, or
        False if not. Raises AttributeError if attrname is not
        a valid attribute for this object.
        """
        if attrname not in self.obj_fields:
            raise AttributeError(
                _("%(objname)s object has no attribute '%(attrname)s'")
                % {'objname': self.obj_name(), 'attrname': attrname}
            )
        return hasattr(self, _get_attrname(attrname))

    @property
    def obj_fields(self) -> list[str]:
        return list(self.fields.keys()) + list(self.obj_extra_fields)

    @property
    def obj_context(self) -> Any:
        return self._context


class ComparableVersionedObject:
    """Mix-in to provide comparison methods

    When objects are to be compared with each other (in tests for example),
    this mixin can be used.
    """

    def __eq__(self, obj: object) -> bool:
        # FIXME(inc0): this can return incorrect value if we consider partially
        # loaded objects from db and fields which are dropped out differ
        if hasattr(obj, 'obj_to_primitive'):
            # This mixin is always combined with VersionedObject
            return bool(
                cast(VersionedObject, self).obj_to_primitive()
                == obj.obj_to_primitive()
            )
        return NotImplemented

    def __hash__(self) -> int:
        return super().__hash__()

    def __ne__(self, obj: object) -> bool:
        if hasattr(obj, 'obj_to_primitive'):
            # This mixin is always combined with VersionedObject
            return bool(
                cast(VersionedObject, self).obj_to_primitive()
                != obj.obj_to_primitive()
            )
        return NotImplemented


class VersionedObjectMixin:
    """Marker base class for OVO mixin classes that contribute fields.

    Mixin classes that define an OVO ``fields`` dict should inherit from this
    class so that the ``oslo_versionedobjects.mypy`` plugin can discover and
    type-check their fields in the same way it handles registered
    ``VersionedObject`` subclasses.

    Example::

        class MyFieldsMixin(VersionedObjectMixin):
            fields: MutableMapping[str, fields.Field[Any]] = {
                'name': fields.StringField(),
            }


        @base.VersionedObjectRegistry.register
        class MyObject(MyFieldsMixin, base.VersionedObject):
            VERSION = '1.0'
            fields = {
                'count': fields.IntegerField(),
            }
    """


class TimestampedObject(VersionedObjectMixin):
    """Mixin class for db backed objects with timestamp fields.

    Sqlalchemy models that inherit from the oslo_db TimestampMixin will include
    these fields and the corresponding objects will benefit from this mixin.
    """

    # Explicitly annotated to match VersionedObject.fields, which allows
    # subclasses to override with specific field dicts without type errors.
    fields: MutableMapping[str, obj_fields.Field[Any]] = {
        'created_at': obj_fields.DateTimeField(nullable=True),
        'updated_at': obj_fields.DateTimeField(nullable=True),
    }


class VersionedObjectDictCompat:
    """Mix-in to provide dictionary key access compatibility

    If an object needs to support attribute access using
    dictionary items instead of object attributes, inherit
    from this class. This should only be used as a temporary
    measure until all callers are converted to use modern
    attribute access.
    """

    def __iter__(self) -> Iterator[str]:
        # This mixin is always combined with VersionedObject
        vo = cast(VersionedObject, self)
        for name in vo.obj_fields:
            if vo.obj_attr_is_set(name) or name in vo.obj_extra_fields:
                yield name

    keys = __iter__

    def values(self) -> Iterator[Any]:
        for name in self:
            yield getattr(self, name)

    def items(self) -> Iterator[tuple[str, Any]]:
        for name in self:
            yield name, getattr(self, name)

    def __getitem__(self, name: str) -> Any:
        return getattr(self, name)

    def __setitem__(self, name: str, value: Any) -> None:
        setattr(self, name, value)

    def get(self, key: str, value: Any = _NotSpecifiedSentinel) -> Any:
        # This mixin is always combined with VersionedObject
        vo = cast(VersionedObject, self)
        if key not in vo.obj_fields:
            raise AttributeError(
                f"'{self.__class__}' object has no attribute '{key}'"
            )
        if value != _NotSpecifiedSentinel and not vo.obj_attr_is_set(key):
            return value
        else:
            return getattr(self, key)

    def update(self, updates: dict[str, Any]) -> None:
        for key, value in updates.items():
            setattr(self, key, value)


class ObjectListBase(Sequence[_VO]):
    """Mixin class for lists of objects.

    This mixin class can be added as a base class for an object that
    is implementing a list of objects. It adds a single field of 'objects',
    which is the list store, and behaves like a list itself. It supports
    serialization of the list of objects automatically.
    """

    _context: Any

    # Explicitly annotated to match VersionedObject.fields, which allows
    # subclasses to override with specific field dicts without type errors.
    fields: MutableMapping[str, obj_fields.Field[Any]] = {
        'objects': obj_fields.ListOfObjectsField('VersionedObject'),
    }

    # This is a dictionary of my_version:child_version mappings so that
    # we can support backleveling our contents based on the version
    # requested of the list object.
    child_versions: Mapping[str, str] = {}

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        if 'objects' not in kwargs:
            self.objects: list[_VO] = []
            # This mixin is always combined with VersionedObject
            cast(VersionedObject, self)._changed_fields.discard('objects')

    def __len__(self) -> int:
        """List length."""
        return len(self.objects)

    def __contains__(self, value: object) -> bool:
        """Check if an object is in the list."""
        return value in self.objects

    @overload
    def __getitem__(self, index: int) -> _VO: ...

    @overload
    def __getitem__(self, index: slice) -> ObjectListBase[_VO]: ...

    def __getitem__(self, index: int | slice) -> _VO | ObjectListBase[_VO]:
        """List index access."""
        if isinstance(index, slice):
            new_obj = self.__class__()
            new_obj.objects = self.objects[index]
            # NOTE(danms): We must be mixed in with a VersionedObject!
            cast(VersionedObject, new_obj).obj_reset_changes()
            new_obj._context = self._context
            return new_obj
        return self.objects[index]

    def sort(
        self,
        key: Callable[[_VO], Any] | None = None,
        reverse: bool = False,
    ) -> None:
        self.objects.sort(key=key, reverse=reverse)  # type: ignore[arg-type]

    def obj_make_compatible(
        self, primitive: dict[str, Any], target_version: str
    ) -> None:
        # Give priority to using child_versions, if that isn't set, try
        # obj_relationships
        if self.child_versions:
            relationships: Any = self.child_versions.items()
        else:
            try:
                # This mixin is always combined with VersionedObject
                relationships = cast(
                    VersionedObject, self
                )._obj_relationship_for('objects', target_version)
            except exception.ObjectActionError:
                # No relationship for this found in manifest or
                # in obj_relationships
                relationships = {}

        try:
            # NOTE(rlrossit): If we have no version information, just
            # backport to child version 1.0 (maintaining default
            # behavior)
            if relationships:
                _get_subobject_version(
                    target_version,
                    relationships,
                    lambda ver: _do_subobject_backport(
                        ver,
                        self,
                        'objects',
                        primitive,
                    ),
                )
            else:
                _do_subobject_backport('1.0', self, 'objects', primitive)
        except exception.TargetBeforeSubobjectExistedException:
            # Child did not exist, so delete it from the primitive
            del primitive['objects']

    def obj_what_changed(self) -> set[str]:
        # NOTE: This mixin expects to be combined with VersionedObject
        # This mixin is always combined with VersionedObject
        changes = set(cast(VersionedObject, self)._changed_fields)
        for child in self.objects:
            if child.obj_what_changed():
                changes.add('objects')
        return changes

    def __add__(self, other: ObjectListBase[_VO]) -> ObjectListBase[_VO]:
        # Handling arbitrary fields may not make sense if those fields are not
        # all concatenatable. Only concatenate if the base 'objects' field is
        # the only one and the classes match.
        if self.__class__ == other.__class__ and list(
            self.__class__.fields.keys()
        ) == ['objects']:
            return self.__class__(objects=self.objects + other.objects)
        else:
            raise TypeError(
                "List Objects should be of the same type and only "
                "have an 'objects' field"
            )

    def __radd__(self, other: ObjectListBase[_VO]) -> ObjectListBase[_VO]:
        if self.__class__ == other.__class__ and list(
            self.__class__.fields.keys()
        ) == ['objects']:
            # This should never be run in practice. If the above condition is
            # met then __add__ would have been run.
            raise NotImplementedError(
                '__radd__ is not implemented for objects of the same type'
            )
        else:
            raise TypeError(
                "List Objects should be of the same type and only "
                "have an 'objects' field"
            )


class VersionedObjectSerializer(messaging.NoOpSerializer):  # type: ignore[misc]
    """A VersionedObject-aware Serializer.

    This implements the Oslo Serializer interface and provides the
    ability to serialize and deserialize VersionedObject entities. Any service
    that needs to accept or return VersionedObjects as arguments or result
    values should pass this to its RPCClient and RPCServer objects.
    """

    # Base class to use for object hydration
    OBJ_BASE_CLASS = VersionedObject

    def _do_backport(
        self,
        context: Any,
        objprim: dict[str, Any],
        objclass: type[VersionedObject],
    ) -> VersionedObject:
        obj_versions = obj_tree_get_versions(objclass.obj_name())
        indirection_api = self.OBJ_BASE_CLASS.indirection_api
        assert indirection_api is not None
        try:
            # NOTE: The API signature says VersionedObject but primitive is
            # passed here; the implementation handles both
            return indirection_api.object_backport_versions(
                context,
                cast(VersionedObject, objprim),
                obj_versions,
            )
        except NotImplementedError:
            # FIXME(danms): Maybe start to warn here about deprecation?
            return indirection_api.object_backport(
                context,
                cast(VersionedObject, objprim),
                objclass.VERSION,
            )

    def _process_object(
        self, context: Any, objprim: dict[str, Any]
    ) -> VersionedObject | None:
        try:
            return self.OBJ_BASE_CLASS.obj_from_primitive(
                objprim, context=context
            )
        except exception.IncompatibleObjectVersion:
            with excutils.save_and_reraise_exception(reraise=False) as ctxt:
                verkey = f'{self.OBJ_BASE_CLASS.OBJ_SERIAL_NAMESPACE}.version'
                objver = objprim[verkey]
                if objver.count('.') == 2:
                    # NOTE(danms): For our purposes, the .z part of the version
                    # should be safe to accept without requiring a backport
                    objprim[verkey] = '.'.join(objver.split('.')[:2])
                    return self._process_object(context, objprim)
                namekey = f'{self.OBJ_BASE_CLASS.OBJ_SERIAL_NAMESPACE}.name'
                objname = objprim[namekey]
                supported = VersionedObjectRegistry.obj_classes().get(
                    objname, []
                )
                if self.OBJ_BASE_CLASS.indirection_api and supported:
                    return self._do_backport(context, objprim, supported[0])
                else:
                    ctxt.reraise = True
        return None

    def _process_iterable(
        self,
        context: Any,
        action_fn: Callable[[Any, Any], Any],
        values: tuple[Any, ...] | list[Any] | set[Any] | dict[str, Any],
    ) -> tuple[Any, ...] | list[Any] | dict[str, Any]:
        """Process an iterable, taking an action on each value.

        :param context: Request context
        :param action_fn: Action to take on each item in values
        :param values: Iterable container of things to take action on
        :returns: A new container of the same type (except set) with
                  items from values having had action applied.
        """
        if isinstance(values, dict):
            return type(values)(
                [(k, action_fn(context, v)) for k, v in values.items()]
            )
        elif isinstance(values, set):
            # NOTE(danms, gibi) A set can't have an unhashable value inside,
            # such as a dict. Convert the set to list, which is fine, since we
            # can't send them over RPC anyway. We convert it to list as this
            # way there will be no semantic change between the fake rpc driver
            # used in functional test and a normal rpc driver.
            return list([action_fn(context, value) for value in values])
        elif isinstance(values, tuple):
            return tuple([action_fn(context, value) for value in values])
        else:
            # list
            return list([action_fn(context, value) for value in values])

    def serialize_entity(self, context: Any, entity: Any) -> Any:
        if isinstance(entity, (tuple, list, set, dict)):
            entity = self._process_iterable(
                context, self.serialize_entity, entity
            )
        elif hasattr(entity, 'obj_to_primitive') and callable(
            entity.obj_to_primitive
        ):
            entity = entity.obj_to_primitive()
        return entity

    def deserialize_entity(self, context: Any, entity: Any) -> Any:
        namekey = f'{self.OBJ_BASE_CLASS.OBJ_SERIAL_NAMESPACE}.name'
        if isinstance(entity, dict) and namekey in entity:
            entity = self._process_object(context, entity)
        elif isinstance(entity, (tuple, list, set, dict)):
            entity = self._process_iterable(
                context, self.deserialize_entity, entity
            )
        return entity


class VersionedObjectIndirectionAPI(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def object_action(
        self,
        context: Any,
        objinst: VersionedObject,
        objmethod: str,
        args: tuple[Any, ...],
        kwargs: dict[str, Any],
    ) -> tuple[dict[str, Any], Any]:
        """Perform an action on a VersionedObject instance.

        When indirection_api is set on a VersionedObject (to a class
        implementing this interface), method calls on remotable methods
        will cause this to be executed to actually make the desired
        call. This often involves performing RPC.

        :param context: The context within which to perform the action
        :param objinst: The object instance on which to perform the action
        :param objmethod: The name of the action method to call
        :param args: The positional arguments to the action method
        :param kwargs: The keyword arguments to the action method
        :returns: The result of the action method
        """
        ...

    def object_class_action(
        self,
        context: Any,
        objname: str,
        objmethod: str,
        objver: str,
        args: tuple[Any, ...],
        kwargs: dict[str, Any],
    ) -> VersionedObject | Any:
        """.. deprecated:: 0.10.0

        Use :func:`object_class_action_versions` instead.

        Perform an action on a VersionedObject class.

        When indirection_api is set on a VersionedObject (to a class
        implementing this interface), classmethod calls on
        remotable_classmethod methods will cause this to be executed to
        actually make the desired call. This usually involves performing
        RPC.

        :param context: The context within which to perform the action
        :param objname: The registry name of the object
        :param objmethod: The name of the action method to call
        :param objver: The (remote) version of the object on which the
                       action is being taken
        :param args: The positional arguments to the action method
        :param kwargs: The keyword arguments to the action method
        :returns: The result of the action method, which may (or may not)
                  be an instance of the implementing VersionedObject class.
        """
        pass

    def object_class_action_versions(
        self,
        context: Any,
        objname: str,
        objmethod: str,
        object_versions: dict[str, str],
        args: tuple[Any, ...],
        kwargs: dict[str, Any],
    ) -> VersionedObject | Any:
        """Perform an action on a VersionedObject class.

        When indirection_api is set on a VersionedObject (to a class
        implementing this interface), classmethod calls on
        remotable_classmethod methods will cause this to be executed to
        actually make the desired call. This usually involves performing
        RPC.

        This differs from object_class_action() in that it is provided
        with object_versions, a manifest of client-side object versions
        for easier nested backports. The manifest is the result of
        calling obj_tree_get_versions().

        NOTE: This was not in the initial spec for this interface, so the
        base class raises NotImplementedError if you don't implement it.
        For backports, this method will be tried first, and if unimplemented,
        will fall back to object_class_action(). New implementations should
        provide this method instead of object_class_action()

        :param context: The context within which to perform the action
        :param objname: The registry name of the object
        :param objmethod: The name of the action method to call
        :param object_versions: A dict of {objname: version} mappings
        :param args: The positional arguments to the action method
        :param kwargs: The keyword arguments to the action method
        :returns: The result of the action method, which may (or may not)
                  be an instance of the implementing VersionedObject class.
        """
        warnings.warn(
            'object_class_action() is deprecated in favor of '
            'object_class_action_versions() and will be removed '
            'in a later release',
            DeprecationWarning,
        )
        raise NotImplementedError('Multi-version class action not supported')

    def object_backport(
        self, context: Any, objinst: VersionedObject, target_version: str
    ) -> VersionedObject:
        """.. deprecated:: 0.10.0

        Use :func:`object_backport_versions` instead.

        Perform a backport of an object instance to a specified version.

        When indirection_api is set on a VersionedObject (to a class
        implementing this interface), the default behavior of the base
        VersionedObjectSerializer, upon receiving an object with a version
        newer than what is in the lcoal registry, is to call this method to
        request a backport of the object. In an environment where there is
        an RPC-able service on the bus which can gracefully downgrade newer
        objects for older services, this method services as a translation
        mechanism for older code when receiving objects from newer code.

        NOTE: This older/original method is soon to be deprecated. When a
        backport is required, the newer object_backport_versions() will be
        tried, and if it raises NotImplementedError, then we will fall back
        to this (less optimal) method.

        :param context: The context within which to perform the backport
        :param objinst: An instance of a VersionedObject to be backported
        :param target_version: The maximum version of the objinst's class
                               that is understood by the requesting host.
        :returns: The downgraded instance of objinst
        """
        raise NotImplementedError('Backport not supported')

    def object_backport_versions(
        self,
        context: Any,
        objinst: VersionedObject,
        object_versions: dict[str, str],
    ) -> VersionedObject:
        """Perform a backport of an object instance.

        This method is basically just like object_backport() but instead of
        providing a specific target version for the toplevel object and
        relying on the service-side mapping to handle sub-objects, this sends
        a mapping of all the dependent objects and their client-supported
        versions. The server will backport objects within the tree starting
        at objinst to the versions specified in object_versions, removing
        objects that have no entry. Use obj_tree_get_versions() to generate
        this mapping.

        NOTE: This was not in the initial spec for this interface, so the
        base class raises NotImplementedError if you don't implement it.
        For backports, this method will be tried first, and if unimplemented,
        will fall back to object_backport().

        :param context: The context within which to perform the backport
        :param objinst: An instance of a VersionedObject to be backported
        :param object_versions: A dict of {objname: version} mappings
        """
        warnings.warn(
            'object_backport() is deprecated in favor of '
            'object_backport_versions() and will be removed '
            'in a later release',
            DeprecationWarning,
        )
        raise NotImplementedError('Multi-version backport not supported')


def obj_make_list(
    context: Any,
    list_obj: _VOL,
    item_cls: type[_VO],
    db_list: list[dict[str, Any]],
    **extra_args: Any,
) -> _VOL:
    """Construct an object list from a list of primitives.

    This calls item_cls._from_db_object() on each item of db_list, and
    adds the resulting object to list_obj.

    :param context: Request context
    :param list_obj: An ObjectListBase object
    :param item_cls: The VersionedObject class of the objects within the list
    :param db_list: The list of primitives to convert to objects
    :param extra_args: Extra arguments to pass to _from_db_object()
    :returns: list_obj
    """
    list_obj.objects = []
    for db_item in db_list:
        # _from_db_object is a convention classmethod for VersionedObject
        # subclasses that work with database backends; it is not declared in
        # the VersionedObject base class since not all objects support it
        item = getattr(item_cls, '_from_db_object')(
            context, item_cls(), db_item, **extra_args
        )
        list_obj.objects.append(item)
    # list_obj is always an ObjectListBase combined with VersionedObject
    list_obj._context = context
    cast(VersionedObject, list_obj).obj_reset_changes()
    return list_obj


def obj_tree_get_versions(
    objname: str, tree: dict[str, str] | None = None
) -> dict[str, str]:
    """Construct a mapping of dependent object versions.

    This method builds a list of dependent object versions given a top-
    level object with other objects as fields. It walks the tree recursively
    to determine all the objects (by symbolic name) that could be contained
    within the top-level object, and the maximum versions of each. The result
    is a dict like::

      {'MyObject': '1.23', ... }

    :param objname: The top-level object at which to start
    :param tree: Used internally, pass None here.
    :returns: A dictionary of object names and versions
    """
    if tree is None:
        tree = {}
    if objname in tree:
        return tree
    objclass = VersionedObjectRegistry.obj_classes()[objname][0]
    tree[objname] = objclass.VERSION
    for field_name in objclass.fields:
        field = objclass.fields[field_name]
        if isinstance(field, obj_fields.ObjectField):
            # ObjectField stores objname directly as an instance attribute
            child_cls = field.objname
        elif isinstance(field, obj_fields.ListOfObjectsField):
            # ListOfObjectsField stores objname directly as an instance
            # attribute
            child_cls = field.objname
        else:
            continue

        try:
            obj_tree_get_versions(child_cls, tree=tree)
        except IndexError:
            raise exception.UnregisteredSubobject(
                child_objname=child_cls, parent_objname=objname
            )
    return tree


def _get_subobject_version(
    tgt_version: str,
    relationships: list[tuple[str, str]] | Any,
    backport_func: Callable[[str], None],
) -> None:
    """Get the version to which we need to convert a subobject.

    This uses the relationships between a parent and a subobject,
    along with the target parent version, to decide the version we need
    to convert a subobject to. If the subobject did not exist in the parent at
    the target version, TargetBeforeChildExistedException is raised. If there
    is a need to backport, backport_func is called and the subobject version
    to backport to is passed in.

    :param tgt_version: The version we are converting the parent to
    :param relationships: A list of (parent, subobject) version tuples
    :param backport_func: A backport function that takes in the subobject
                          version
    :returns: The version we need to convert the subobject to
    """
    tgt = vutils.convert_version_to_tuple(tgt_version)
    for index, versions in enumerate(relationships):
        parent, child = versions
        parent = vutils.convert_version_to_tuple(parent)
        if tgt < parent:
            if index == 0:
                # We're backporting to a version of the parent that did
                # not contain this subobject
                raise exception.TargetBeforeSubobjectExistedException(
                    target_version=tgt_version
                )
            else:
                # We're in a gap between index-1 and index, so set the desired
                # version to the previous index's version
                child = relationships[index - 1][1]
                backport_func(child)
            return
        elif tgt == parent:
            # We found the version we want, so backport to it
            backport_func(child)
            return


def _do_subobject_backport(
    to_version: str,
    parent: VersionedObject | ObjectListBase[Any],
    field: str,
    primitive: dict[str, Any],
) -> None:
    obj = getattr(parent, field)
    manifest = (
        hasattr(parent, '_obj_version_manifest')
        and parent._obj_version_manifest
        or None
    )
    if isinstance(obj, VersionedObject):
        obj.obj_make_compatible_from_manifest(
            obj._obj_primitive_field(primitive[field], 'data'),
            to_version,
            version_manifest=manifest,
        )
        ver_key = obj._obj_primitive_key('version')
        primitive[field][ver_key] = to_version
    elif isinstance(obj, list):
        for i, element in enumerate(obj):
            element.obj_make_compatible_from_manifest(
                element._obj_primitive_field(primitive[field][i], 'data'),
                to_version,
                version_manifest=manifest,
            )
            ver_key = element._obj_primitive_key('version')
            primitive[field][i][ver_key] = to_version



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_versionedobjects import base
from oslo_versionedobjects import fields


@base.VersionedObjectRegistry.register
class Inner(base.VersionedObject):
    VERSION = '1.0'
    fields = {
        'name': fields.StringField(),
    }


@base.VersionedObjectRegistry.register
class Outer(base.VersionedObject):
    VERSION = '1.0'
    fields = {
        # ObjectField and ListOfObjectsField resolve to the concrete objtype
        # class (not just the VersionedObject base), because the plugin looks
        # up the named class in the mypy symbol table at analysis time.
        'child': fields.ObjectField('Inner'),
        'children': fields.ListOfObjectsField('Inner'),
        'nullable_child': fields.ObjectField('Inner', nullable=True),
    }


obj = Outer()
reveal_type(
    obj.child  # N: Revealed type is "oslo_versionedobjects.tests.test_data.object_fields.Inner"
)
reveal_type(
    obj.children  # N: Revealed type is "list[oslo_versionedobjects.tests.test_data.object_fields.Inner]"
)
reveal_type(
    obj.nullable_child  # N: Revealed type is "oslo_versionedobjects.tests.test_data.object_fields.Inner | None"
)



#    Copyright 2015 IBM Corp.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy
import datetime
import hashlib
import inspect
from typing import Any
from unittest import mock

from oslo_versionedobjects import base
from oslo_versionedobjects import exception
from oslo_versionedobjects import fields
from oslo_versionedobjects import fixture
from oslo_versionedobjects import test


class MyObject(base.VersionedObject):
    fields = {'diglett': fields.IntegerField()}

    @base.remotable
    def remotable_method(self):
        pass

    @classmethod
    @base.remotable
    def remotable_classmethod(cls):
        pass

    def non_remotable_method(self):
        pass

    @classmethod
    def non_remotable_classmethod(cls):
        pass


class MyObject2(base.VersionedObject):
    pass


class MyExtraObject(base.VersionedObject):
    pass


@base.VersionedObjectRegistry.register_if(False)
class MyListObject(base.ObjectListBase[MyObject], base.VersionedObject):
    fields = {'objects': fields.ListOfObjectsField('MyObject')}


class TestObjectComparators(test.TestCase):
    @base.VersionedObjectRegistry.register_if(False)
    class MyComparedObject(base.VersionedObject):
        fields = {'foo': fields.IntegerField(), 'bar': fields.IntegerField()}

    @base.VersionedObjectRegistry.register_if(False)
    class MyComparedObjectWithTZ(base.VersionedObject):
        fields = {'tzfield': fields.DateTimeField()}

    def test_compare_obj(self):
        mock_test = mock.Mock()
        mock_test.assertEqual = mock.Mock()
        my_obj = self.MyComparedObject(foo=1, bar=2)
        my_db_obj = {'foo': 1, 'bar': 2}

        fixture.compare_obj(mock_test, my_obj, my_db_obj)

        expected_calls = [(1, 1), (2, 2)]
        actual_calls = [c[0] for c in mock_test.assertEqual.call_args_list]
        for call in expected_calls:
            self.assertIn(call, actual_calls)

    def test_compare_obj_with_unset(self):
        # If the object has nothing set, and also the db object has the same
        # thing not set, it's OK.
        mock_test = mock.Mock()
        mock_test.assertEqual = mock.Mock()
        my_obj = self.MyComparedObject()

        fixture.compare_obj(mock_test, my_obj, {})

        self.assertFalse(
            mock_test.assertEqual.called,
            "assertEqual should not have been called, there is nothing to "
            "compare.",
        )

    def test_compare_obj_with_unset_in_obj(self):
        # If the db dict has something set, but the object doesn't, that's !=
        mock_test = mock.Mock()
        mock_test.assertEqual = mock.Mock()
        my_obj = self.MyComparedObject(foo=1)
        my_db_obj = {'foo': 1, 'bar': 2}

        self.assertRaises(
            AssertionError, fixture.compare_obj, mock_test, my_obj, my_db_obj
        )

    def test_compare_obj_with_unset_in_db_dict(self):
        # If the object has something set, but the db dict doesn't, that's !=
        mock_test = mock.Mock()
        mock_test.assertEqual = mock.Mock()
        my_obj = self.MyComparedObject(foo=1, bar=2)
        my_db_obj = {'foo': 1}

        self.assertRaises(
            AssertionError, fixture.compare_obj, mock_test, my_obj, my_db_obj
        )

    def test_compare_obj_with_unset_in_obj_ignored(self):
        # If the db dict has something set, but the object doesn't, but we
        # ignore that key, we are equal
        my_obj = self.MyComparedObject(foo=1)
        my_db_obj = {'foo': 1, 'bar': 2}
        ignore = ['bar']

        fixture.compare_obj(self, my_obj, my_db_obj, allow_missing=ignore)

    def test_compare_obj_with_unset_in_db_dict_ignored(self):
        # If the object has something set, but the db dict doesn't, but we
        # ignore that key, we are equal
        my_obj = self.MyComparedObject(foo=1, bar=2)
        my_db_obj = {'foo': 1}
        ignore = ['bar']

        fixture.compare_obj(self, my_obj, my_db_obj, allow_missing=ignore)

    def test_compare_obj_with_allow_missing_unequal(self):
        # If the tested key is in allow_missing, but both the obj and db_obj
        # have the value set, we should still check it for equality
        mock_test = mock.Mock()
        mock_test.assertEqual = mock.Mock()
        my_obj = self.MyComparedObject(foo=1, bar=2)
        my_db_obj = {'foo': 1, 'bar': 1}
        ignore = ['bar']

        fixture.compare_obj(mock_test, my_obj, my_db_obj, allow_missing=ignore)

        expected_calls = [(1, 1), (1, 2)]
        actual_calls = [c[0] for c in mock_test.assertEqual.call_args_list]
        for call in expected_calls:
            self.assertIn(call, actual_calls)

    def test_compare_obj_with_subs(self):
        mock_test = mock.Mock()
        mock_test.assertEqual = mock.Mock()
        my_obj = self.MyComparedObject(foo=1, bar=2)
        my_db_obj = {'doo': 1, 'bar': 2}
        subs = {'foo': 'doo'}

        fixture.compare_obj(mock_test, my_obj, my_db_obj, subs=subs)

        expected_calls = [(1, 1), (2, 2)]
        actual_calls = [c[0] for c in mock_test.assertEqual.call_args_list]
        for call in expected_calls:
            self.assertIn(call, actual_calls)

    def test_compare_obj_with_allow_missing(self):
        mock_test = mock.Mock()
        mock_test.assertEqual = mock.Mock()
        my_obj = self.MyComparedObject(foo=1)
        my_db_obj = {'foo': 1, 'bar': 2}
        ignores = ['bar']

        fixture.compare_obj(
            mock_test, my_obj, my_db_obj, allow_missing=ignores
        )

        mock_test.assertEqual.assert_called_once_with(1, 1)

    def test_compare_obj_with_comparators(self):
        mock_test = mock.Mock()
        mock_test.assertEqual = mock.Mock()
        comparator = mock.Mock()
        comp_dict = {'foo': comparator}
        my_obj = self.MyComparedObject(foo=1, bar=2)
        my_db_obj = {'foo': 1, 'bar': 2}

        fixture.compare_obj(
            mock_test, my_obj, my_db_obj, comparators=comp_dict
        )

        comparator.assert_called_once_with(1, 1)
        mock_test.assertEqual.assert_called_once_with(2, 2)

    def test_compare_obj_with_dt(self):
        mock_test = mock.Mock()
        mock_test.assertEqual = mock.Mock()
        dt = datetime.datetime(1955, 11, 5, tzinfo=datetime.UTC)
        replaced_dt = dt.replace(tzinfo=None)
        my_obj = self.MyComparedObjectWithTZ(tzfield=dt)
        my_db_obj = {'tzfield': replaced_dt}

        fixture.compare_obj(mock_test, my_obj, my_db_obj)

        mock_test.assertEqual.assert_called_once_with(replaced_dt, replaced_dt)


class FakeResource(base.VersionedObject):
    # Version 1.0: Initial version
    VERSION = '1.0'

    fields = {'identifier': fields.Field(fields.Integer(), default=123)}


class TestObjectVersionChecker(test.TestCase):
    def setUp(self):
        super().setUp()
        objects = [
            MyObject,
            MyObject2,
        ]
        self.obj_classes = {obj.__name__: [obj] for obj in objects}
        self.ovc = fixture.ObjectVersionChecker(obj_classes=self.obj_classes)

    def test_get_hashes(self):
        # Make sure get_hashes retrieves the fingerprint of all objects
        fp = 'ashketchum'
        with mock.patch.object(self.ovc, '_get_fingerprint') as mock_gf:
            mock_gf.return_value = fp
            actual = self.ovc.get_hashes()

        expected = self._generate_hashes(self.obj_classes, fp)
        self.assertEqual(
            expected,
            actual,
            "ObjectVersionChecker is not getting the fingerprints of all "
            "registered objects.",
        )

    def test_get_hashes_with_extra_data(self):
        # Make sure get_hashes uses the extra_data_func
        fp = 'garyoak'
        mock_func = mock.MagicMock()
        with mock.patch.object(self.ovc, '_get_fingerprint') as mock_gf:
            mock_gf.return_value = fp
            actual = self.ovc.get_hashes(extra_data_func=mock_func)

        expected = self._generate_hashes(self.obj_classes, fp)
        expected_calls = [
            ((name,), {'extra_data_func': mock_func})
            for name in self.obj_classes.keys()
        ]

        self.assertEqual(
            expected,
            actual,
            "ObjectVersionChecker is not getting the fingerprints of all "
            "registered objects.",
        )

        self.assertEqual(
            len(expected_calls),
            len(mock_gf.call_args_list),
            "get_hashes() did not call get the fingerprints of "
            "all objects in the registry.",
        )
        for call in expected_calls:
            self.assertIn(
                call,
                mock_gf.call_args_list,
                "get_hashes() did not call _get_fingerprint()correctly.",
            )

    def test_test_hashes_none_changed(self):
        # Make sure test_hashes() generates an empty dictionary when
        # there are no objects that have changed
        fp = 'pikachu'
        hashes = self._generate_hashes(self.obj_classes, fp)

        with mock.patch.object(self.ovc, 'get_hashes') as mock_gh:
            mock_gh.return_value = hashes
            # I'm so sorry, but they have to be named this way
            actual_expected, actual_actual = self.ovc.test_hashes(hashes)

        self.assertEqual(
            {},
            actual_expected,
            "There are no objects changed, so the 'expected' return value "
            "should contain no objects.",
        )
        self.assertEqual(
            {},
            actual_actual,
            "There are no objects changed, so the 'actual' return value "
            "should contain no objects.",
        )

    def test_test_hashes_class_not_added(self):
        # Make sure the expected and actual values differ when a class
        # was added to the registry, but not the static dictionary
        fp = 'gyrados'
        new_classes = copy.copy(self.obj_classes)
        self._add_class(new_classes, MyExtraObject)

        expected_hashes = self._generate_hashes(self.obj_classes, fp)
        actual_hashes = self._generate_hashes(new_classes, fp)

        with mock.patch.object(self.ovc, 'get_hashes') as mock_gh:
            mock_gh.return_value = actual_hashes
            actual_exp, actual_act = self.ovc.test_hashes(expected_hashes)

        expected_expected = {MyExtraObject.__name__: None}
        expected_actual = {MyExtraObject.__name__: fp}

        self.assertEqual(
            expected_expected,
            actual_exp,
            "Expected hashes should not contain the fingerprint of the class "
            "that has not been added to the expected hash dictionary.",
        )
        self.assertEqual(
            expected_actual,
            actual_act,
            "The actual hash should contain the class that was added to the "
            "registry.",
        )

    def test_test_hashes_new_fp_incorrect(self):
        # Make sure the expected and actual values differ when a fingerprint
        # was changed, but the static dictionary was not updated
        fp1 = 'beedrill'
        fp2 = 'snorlax'
        expected_hashes = self._generate_hashes(self.obj_classes, fp1)
        actual_hashes = copy.copy(expected_hashes)

        actual_hashes[MyObject.__name__] = fp2

        with mock.patch.object(self.ovc, 'get_hashes') as mock_gh:
            mock_gh.return_value = actual_hashes
            actual_exp, actual_act = self.ovc.test_hashes(expected_hashes)

        expected_expected = {MyObject.__name__: fp1}
        expected_actual = {MyObject.__name__: fp2}

        self.assertEqual(
            expected_expected,
            actual_exp,
            "Expected hashes should contain the updated object with the old "
            "hash.",
        )
        self.assertEqual(
            expected_actual,
            actual_act,
            "Actual hashes should contain the updated object with the new "
            "hash.",
        )

    def test_test_hashes_passes_extra_func(self):
        # Make sure that test_hashes passes the extra_func to get_hashes
        mock_extra_func = mock.Mock()

        with mock.patch.object(self.ovc, 'get_hashes') as mock_get_hashes:
            self.ovc.test_hashes({}, extra_data_func=mock_extra_func)

            mock_get_hashes.assert_called_once_with(
                extra_data_func=mock_extra_func
            )

    def test_get_dependency_tree(self):
        # Make sure get_dependency_tree() gets the dependencies of all
        # objects in the registry
        with mock.patch.object(self.ovc, '_get_dependencies') as mock_gd:
            self.ovc.get_dependency_tree()

        expected_calls: list[Any] = [(({}, MyObject),), (({}, MyObject2),)]

        self.assertEqual(
            2,
            len(mock_gd.call_args_list),
            "get_dependency_tree() tried to get the dependencies too many "
            "times.",
        )

        for call in expected_calls:
            self.assertIn(
                call,
                mock_gd.call_args_list,
                "get_dependency_tree() did not get the dependencies of the "
                "objects correctly.",
            )

    def test_test_relationships_none_changed(self):
        # Make sure test_relationships() generates an empty dictionary when
        # no relationships have been changed
        dep_tree: dict[str, Any] = {}
        # tree will be {'MyObject': {'MyObject2': '1.0'}}
        self._add_dependency(MyObject, MyObject2, dep_tree)

        with mock.patch.object(self.ovc, 'get_dependency_tree') as mock_gdt:
            mock_gdt.return_value = dep_tree
            actual_exp, actual_act = self.ovc.test_relationships(dep_tree)

        self.assertEqual(
            {},
            actual_exp,
            "There are no objects changed, so the 'expected' return value "
            "should contain no objects.",
        )
        self.assertEqual(
            {},
            actual_act,
            "There are no objects changed, so the 'actual' return value "
            "should contain no objects.",
        )

    def test_test_relationships_rel_added(self):
        # Make sure expected and actual relationships differ if a
        # relationship is added to a class
        exp_tree: dict[str, Any] = {}
        actual_tree: dict[str, Any] = {}
        self._add_dependency(MyObject, MyObject2, exp_tree)
        self._add_dependency(MyObject, MyObject2, actual_tree)
        self._add_dependency(MyObject, MyExtraObject, actual_tree)

        with mock.patch.object(self.ovc, 'get_dependency_tree') as mock_gdt:
            mock_gdt.return_value = actual_tree
            actual_exp, actual_act = self.ovc.test_relationships(exp_tree)

        expected_expected = {'MyObject': {'MyObject2': '1.0'}}
        expected_actual = {
            'MyObject': {'MyObject2': '1.0', 'MyExtraObject': '1.0'}
        }

        self.assertEqual(
            expected_expected,
            actual_exp,
            "The expected relationship tree is not being built from changes "
            "correctly.",
        )
        self.assertEqual(
            expected_actual,
            actual_act,
            "The actual relationship tree is not being built from changes "
            "correctly.",
        )

    def test_test_relationships_class_added(self):
        # Make sure expected and actual relationships differ if a new
        # class is added to the relationship tree
        exp_tree: dict[str, Any] = {}
        actual_tree: dict[str, Any] = {}
        self._add_dependency(MyObject, MyObject2, exp_tree)
        self._add_dependency(MyObject, MyObject2, actual_tree)
        self._add_dependency(MyObject2, MyExtraObject, actual_tree)

        with mock.patch.object(self.ovc, 'get_dependency_tree') as mock_gdt:
            mock_gdt.return_value = actual_tree
            actual_exp, actual_act = self.ovc.test_relationships(exp_tree)

        expected_expected = {'MyObject2': None}
        expected_actual = {'MyObject2': {'MyExtraObject': '1.0'}}

        self.assertEqual(
            expected_expected,
            actual_exp,
            "The expected relationship tree is not being built from changes "
            "correctly.",
        )
        self.assertEqual(
            expected_actual,
            actual_act,
            "The actual relationship tree is not being built from changes "
            "correctly.",
        )

    def test_test_compatibility_routines(self):
        # Make sure test_compatibility_routines() checks the object
        # compatibility of all objects in the registry
        del self.ovc.obj_classes[MyObject2.__name__]

        with mock.patch.object(self.ovc, '_test_object_compatibility') as toc:
            self.ovc.test_compatibility_routines()

        toc.assert_called_once_with(
            MyObject, manifest=None, init_args=[], init_kwargs={}
        )

    def test_test_compatibility_routines_with_manifest(self):
        # Make sure test_compatibility_routines() uses the version manifest
        del self.ovc.obj_classes[MyObject2.__name__]
        man = {'who': 'cares'}

        with mock.patch.object(self.ovc, '_test_object_compatibility') as toc:
            with mock.patch(
                'oslo_versionedobjects.base.obj_tree_get_versions'
            ) as otgv:
                otgv.return_value = man
                self.ovc.test_compatibility_routines(use_manifest=True)

        otgv.assert_called_once_with(MyObject.__name__)
        toc.assert_called_once_with(
            MyObject, manifest=man, init_args=[], init_kwargs={}
        )

    def test_test_compatibility_routines_with_args_kwargs(self):
        # Make sure test_compatibility_routines() uses init args/kwargs
        del self.ovc.obj_classes[MyObject2.__name__]

        with mock.patch.object(self.ovc, '_test_object_compatibility') as toc:
            self.ovc.test_compatibility_routines(
                init_args={MyObject: [1]},
                init_kwargs={MyObject: {'foo': 'bar'}},
            )

        toc.assert_called_once_with(
            MyObject, manifest=None, init_args=[1], init_kwargs={'foo': 'bar'}
        )

    def test_test_relationships_in_order(self):
        # Make sure test_relationships_in_order() tests the relationships
        # of all objects in the registry
        with mock.patch.object(
            self.ovc, '_test_relationships_in_order'
        ) as mock_tr:
            self.ovc.test_relationships_in_order()

        expected_calls = [((MyObject,),), ((MyObject2,),)]

        self.assertEqual(
            2,
            len(mock_tr.call_args_list),
            "test_relationships_in_order() tested too many relationships.",
        )
        for call in expected_calls:
            self.assertIn(
                call,
                mock_tr.call_args_list,
                "test_relationships_in_order() did not test the "
                "relationships of the individual objects "
                "correctly.",
            )

    def test_test_relationships_in_order_positive(self):
        # Make sure a correct relationship ordering doesn't blow up
        rels = {'bellsprout': [('1.0', '1.0'), ('1.1', '1.2'), ('1.3', '1.3')]}
        MyObject.obj_relationships = rels

        self.ovc._test_relationships_in_order(MyObject)

    def test_test_relationships_in_order_negative(self):
        # Make sure an out-of-order relationship does blow up
        rels = {'rattata': [('1.0', '1.0'), ('1.1', '1.2'), ('1.3', '1.1')]}
        MyObject.obj_relationships = rels

        self.assertRaises(
            AssertionError, self.ovc._test_relationships_in_order, MyObject
        )

    def test_find_remotable_method(self):
        # Make sure we can find a remotable method on an object
        method = self.ovc._find_remotable_method(
            MyObject, MyObject.remotable_method
        )

        self.assertEqual(
            getattr(MyObject.remotable_method, 'original_fn'),
            method,
            "_find_remotable_method() did not find the remotable "
            "method of MyObject.",
        )

    def test_find_remotable_method_classmethod(self):
        # Make sure we can find a remotable classmethod on an object
        rcm = MyObject.remotable_classmethod
        method = self.ovc._find_remotable_method(MyObject, rcm)

        expected = rcm.__get__(None, MyObject).original_fn
        self.assertEqual(
            expected,
            method,
            "_find_remotable_method() did not find the remotable classmethod.",
        )

    def test_find_remotable_method_non_remotable_method(self):
        # Make sure nothing is found when we have only a non-remotable method
        nrm = MyObject.non_remotable_method
        method = self.ovc._find_remotable_method(MyObject, nrm)

        self.assertIsNone(
            method,
            "_find_remotable_method() found a method that isn't remotable.",
        )

    def test_find_remotable_method_non_remotable_classmethod(self):
        # Make sure we don't find a non-remotable classmethod
        nrcm = MyObject.non_remotable_classmethod
        method = self.ovc._find_remotable_method(MyObject, nrcm)

        self.assertIsNone(
            method,
            "_find_remotable_method() found a method that isn't remotable.",
        )

    def test_get_fingerprint(self):
        # Make sure _get_fingerprint() generates a consistent fingerprint
        MyObject.VERSION = '1.1'
        argspec = 'vulpix'

        with mock.patch.object(fixture, 'get_method_spec') as mock_gas:
            mock_gas.return_value = argspec
            fp = self.ovc._get_fingerprint(MyObject.__name__)

        exp_fields = sorted(list(MyObject.fields.items()))
        exp_methods = sorted(
            [('remotable_method', argspec), ('remotable_classmethod', argspec)]
        )
        expected_relevant_data = (exp_fields, exp_methods)
        expected_hash = hashlib.md5(
            bytes(repr(expected_relevant_data).encode()), usedforsecurity=False
        ).hexdigest()
        expected_fp = f'{MyObject.VERSION}-{expected_hash}'

        self.assertEqual(
            expected_fp,
            fp,
            "_get_fingerprint() did not generate a correct fingerprint.",
        )

    def test_get_fingerprint_with_child_versions(self):
        # Make sure _get_fingerprint() generates a consistent fingerprint
        # when child_versions are present
        child_versions = {'1.0': '1.0', '1.1': '1.1'}
        MyObject.child_versions = child_versions  # type: ignore[attr-defined]
        argspec = 'onix'

        with mock.patch.object(fixture, 'get_method_spec') as mock_gas:
            mock_gas.return_value = argspec
            fp = self.ovc._get_fingerprint(MyObject.__name__)

        exp_fields = sorted(list(MyObject.fields.items()))
        exp_methods = sorted(
            [('remotable_method', argspec), ('remotable_classmethod', argspec)]
        )
        exp_child_versions = fixture.OsloOrderedDict(
            sorted(child_versions.items())
        )
        exp_relevant_data = (exp_fields, exp_methods, exp_child_versions)

        expected_hash = hashlib.md5(
            bytes(repr(exp_relevant_data).encode()), usedforsecurity=False
        ).hexdigest()
        expected_fp = f'{MyObject.VERSION}-{expected_hash}'

        self.assertEqual(
            expected_fp,
            fp,
            "_get_fingerprint() did not generate a correct fingerprint.",
        )

    def test_get_fingerprint_with_extra_data(self):
        # Make sure _get_fingerprint() uses extra_data_func when it is
        # supplied
        class ExtraDataObj(base.VersionedObject):
            pass

        def get_data(obj_class):
            return (obj_class,)

        ExtraDataObj.VERSION = '1.1'
        argspec = 'cubone'
        self._add_class(self.obj_classes, ExtraDataObj)

        with mock.patch.object(fixture, 'get_method_spec') as mock_gas:
            mock_gas.return_value = argspec
            fp = self.ovc._get_fingerprint(
                ExtraDataObj.__name__, extra_data_func=get_data
            )

        expected_hash = hashlib.md5(
            bytes(repr(([], [], ExtraDataObj)).encode()),
            usedforsecurity=False,
        ).hexdigest()
        expected_fp = f'{ExtraDataObj.VERSION}-{expected_hash}'

        self.assertEqual(
            expected_fp,
            fp,
            "_get_fingerprint() did not generate a correct fingerprint.",
        )

    def test_get_fingerprint_with_defaulted_set(self):
        class ClassWithDefaultedSetField(base.VersionedObject):
            VERSION = '1.0'
            fields = {
                'empty_default': fields.SetOfIntegersField(default=set()),
                'non_empty_default': fields.SetOfIntegersField(default={1, 2}),
            }

        self._add_class(self.obj_classes, ClassWithDefaultedSetField)

        # it is expected that this hash is stable across python versions
        expected = '1.0-bcc44920f2f727eca463c6eb4fe8445b'
        actual = self.ovc._get_fingerprint(ClassWithDefaultedSetField.__name__)
        self.assertEqual(expected, actual)

    def test_get_dependencies(self):
        # Make sure _get_dependencies() generates a correct tree when parsing
        # an object
        self._add_class(self.obj_classes, MyExtraObject)
        MyObject.fields['subob'] = fields.ObjectField('MyExtraObject')
        MyExtraObject.VERSION = '1.0'
        tree: dict[str, dict[str, str]] = {}

        self.ovc._get_dependencies(tree, MyObject)

        expected_tree = {'MyObject': {'MyExtraObject': '1.0'}}

        self.assertEqual(
            expected_tree,
            tree,
            "_get_dependencies() did not generate a correct dependency tree.",
        )

    def test_test_object_compatibility(self):
        # Make sure _test_object_compatibility() tests obj_to_primitive()
        # on each prior version to the current version
        to_prim = mock.MagicMock(spec=callable)
        MyObject.VERSION = '1.1'
        MyObject.obj_to_primitive = to_prim  # type: ignore[method-assign]

        self.ovc._test_object_compatibility(MyObject)

        expected_calls = [
            ((), {'target_version': '1.0'}),
            ((), {'target_version': '1.1'}),
        ]

        self.assertEqual(
            expected_calls,
            to_prim.call_args_list,
            "_test_object_compatibility() did not test obj_to_primitive() on "
            "the correct target versions",
        )

    def test_test_object_compatibility_args_kwargs(self):
        # Make sure _test_object_compatibility() tests obj_to_primitive()
        # with the correct args and kwargs to init
        to_prim = mock.MagicMock(spec=callable)
        MyObject.obj_to_primitive = to_prim  # type: ignore[method-assign]
        MyObject.VERSION = '1.1'
        args = [1]
        kwargs = {'foo': 'bar'}

        with mock.patch.object(
            MyObject, '__init__', return_value=None
        ) as mock_init:
            self.ovc._test_object_compatibility(
                MyObject, init_args=args, init_kwargs=kwargs
            )

        expected_init = ((1,), {'foo': 'bar'})
        expected_init_calls = [expected_init, expected_init]
        self.assertEqual(
            expected_init_calls,
            mock_init.call_args_list,
            "_test_object_compatibility() did not call __init__() properly "
            "on the object",
        )

        expected_to_prim = [
            ((), {'target_version': '1.0'}),
            ((), {'target_version': '1.1'}),
        ]
        self.assertEqual(
            expected_to_prim,
            to_prim.call_args_list,
            "_test_object_compatibility() did not test obj_to_primitive() "
            "on the correct target versions",
        )

    def _add_class(self, obj_classes, cls):
        obj_classes[cls.__name__] = [cls]

    def _generate_hashes(self, classes, fp):
        # Generate hashes for classes, giving fp as the fingerprint
        # for all classes
        return {cls: fp for cls in classes.keys()}

    def _add_dependency(self, parent_cls, child_cls, tree):
        # Add a dependency to the tree with the parent class holding
        # version 1.0 of the given child class
        deps = tree.get(parent_cls.__name__, {})
        deps[child_cls.__name__] = '1.0'
        tree[parent_cls.__name__] = deps


class TestVersionedObjectRegistryFixture(test.TestCase):
    primitive = {
        'versioned_object.name': 'FakeResource',
        'versioned_object.namespace': 'versionedobjects',
        'versioned_object.version': '1.0',
        'versioned_object.data': {'identifier': 123},
    }

    def test_object_registered_temporarily(self):
        # Test object that has not been registered
        self.assertRaises(
            exception.UnsupportedObjectError,
            FakeResource.obj_from_primitive,
            self.primitive,
        )

        with fixture.VersionedObjectRegistryFixture() as obj_registry:
            # Register object locally
            obj_registry.setUp()
            obj_registry.register(FakeResource)

            # Test object has now been registered
            obj = FakeResource.obj_from_primitive(self.primitive)
            self.assertEqual(obj.identifier, 123)
            self.assertEqual('1.0', obj.VERSION)

        # Test object that is no longer registered
        self.assertRaises(
            exception.UnsupportedObjectError,
            FakeResource.obj_from_primitive,
            self.primitive,
        )


class TestStableObjectJsonFixture(test.TestCase):
    def test_changes_sort(self):
        @base.VersionedObjectRegistry.register_if(False)
        class TestObject(base.VersionedObject):
            fields = {'z': fields.StringField(), 'a': fields.StringField()}

            def obj_what_changed(self):
                return ['z', 'a']

        obj = TestObject(a='foo', z='bar')
        self.assertEqual(
            ['z', 'a'], obj.obj_to_primitive()['versioned_object.changes']
        )
        with fixture.StableObjectJsonFixture():
            self.assertEqual(
                ['a', 'z'], obj.obj_to_primitive()['versioned_object.changes']
            )


class TestMethodSpec(test.TestCase):
    def setUp(self):
        super().setUp()

        def test_method1(a, b, kw1=123, **kwargs):
            pass

        def test_method2(a, b, *args):
            pass

        def test_method3(a, b, *args, kw1=123, **kwargs):
            pass

        self._test_method1 = test_method1
        self._test_method2 = test_method2
        self._test_method3 = test_method3

    def test_method_spec_compat(self):
        self.assertEqual(
            fixture.ArgSpec(
                args=['a', 'b', 'kw1'],
                varargs=None,
                keywords='kwargs',
                defaults=(123,),
            ),
            fixture.get_method_spec(self._test_method1),
        )
        self.assertEqual(
            fixture.ArgSpec(
                args=['a', 'b'], varargs='args', keywords=None, defaults=None
            ),
            fixture.get_method_spec(self._test_method2),
        )
        self.assertEqual(
            inspect.getfullargspec(self._test_method3),
            fixture.get_method_spec(self._test_method3),
        )



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_versionedobjects import base
from oslo_versionedobjects import fields


@base.VersionedObjectRegistry.register
class BasicFieldsObject(base.VersionedObject):
    VERSION = '1.0'
    fields = {
        'uuid': fields.UUIDField(),
        'name': fields.StringField(),
        'count': fields.IntegerField(),
        'ratio': fields.FloatField(),
        'active': fields.BooleanField(),
        'created_at': fields.DateTimeField(nullable=True),
    }


obj = BasicFieldsObject()
reveal_type(obj.uuid)  # N: Revealed type is "str"
reveal_type(obj.name)  # N: Revealed type is "str"
reveal_type(obj.count)  # N: Revealed type is "int"
reveal_type(obj.ratio)  # N: Revealed type is "float"
reveal_type(obj.active)  # N: Revealed type is "bool"
reveal_type(obj.created_at)  # N: Revealed type is "datetime.datetime | None"



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from unittest import mock

import fixtures
from mypy import nodes
from mypy import options as mypy_options
from mypy import types

from oslo_versionedobjects import mypy as ovo_mypy
from oslo_versionedobjects import test


def _make_plugin():
    opts = mypy_options.Options()
    return ovo_mypy.OsloVersionedObjectPlugin(opts)


def _make_class_info(name, module_name='mymodule'):
    """Create a TypeInfo (class info) with an empty class body."""
    sym_table = nodes.SymbolTable()
    block = nodes.Block([])
    cls_def = nodes.ClassDef(name, block)
    type_info = nodes.TypeInfo(sym_table, cls_def, module_name)
    type_info._fullname = f'{module_name}.{name}'
    cls_def.info = type_info
    return type_info


def _make_fields_assignment(*field_names_and_fullnames):
    """Create an AST AssignmentStmt for ``fields = {...}``.

    Each argument should be a ``(field_name, field_type_fullname)`` tuple.
    """
    items: list[tuple[nodes.Expression | None, nodes.Expression]] = []
    for field_name, field_type_fullname in field_names_and_fullnames:
        key = nodes.StrExpr(field_name)
        callee = nodes.NameExpr(field_type_fullname.split('.')[-1])
        callee.fullname = field_type_fullname
        call = nodes.CallExpr(callee, [], [], [])
        items.append((key, call))
    lvalue = nodes.NameExpr('fields')
    return nodes.AssignmentStmt([lvalue], nodes.DictExpr(items))


def _make_object_field_assignment(
    field_name: str,
    field_type_fullname: str,
    objtype_name: str,
) -> nodes.AssignmentStmt:
    """Create an AST for ``fields = {field_name: ObjectField(objtype_name)}``.

    The ``objtype_name`` is inserted as a positional argument, matching the
    runtime signature of ``ObjectField`` and ``ListOfObjectsField``.
    """
    key = nodes.StrExpr(field_name)
    callee = nodes.NameExpr(field_type_fullname.split('.')[-1])
    callee.fullname = field_type_fullname
    call = nodes.CallExpr(
        callee,
        [nodes.StrExpr(objtype_name)],
        [nodes.ARG_POS],
        [None],
    )
    lvalue = nodes.NameExpr('fields')
    return nodes.AssignmentStmt([lvalue], nodes.DictExpr([(key, call)]))


def _make_field_type(field_python_type):
    """Create an AST for a ``FieldType``."""
    auto_cls_def = nodes.ClassDef('AutoTypedField', nodes.Block([]))
    auto_type_info = nodes.TypeInfo(
        nodes.SymbolTable(), auto_cls_def, 'oslo_versionedobjects.fields'
    )
    auto_type_info._fullname = 'oslo_versionedobjects.fields.AutoTypedField'
    auto_cls_def.info = auto_type_info

    field_cls_def = nodes.ClassDef('IntegerField', nodes.Block([]))
    field_type_info = nodes.TypeInfo(
        nodes.SymbolTable(), field_cls_def, 'oslo_versionedobjects.fields'
    )
    field_type_info._fullname = 'oslo_versionedobjects.fields.IntegerField'
    field_type_info.bases = [
        types.Instance(auto_type_info, [field_python_type])
    ]
    field_type_info.mro = [field_type_info, auto_type_info]
    field_cls_def.info = field_type_info

    return nodes.SymbolTableNode(nodes.GDEF, field_type_info)


def _make_ctx(name='MyObj', statements=None, module_name='mymodule'):
    """Create a mock ClassDefContext with the given class body statements."""
    type_info = _make_class_info(name, module_name)
    type_info.defn.defs.body = statements if statements is not None else []
    ctx = mock.MagicMock()
    ctx.cls = type_info.defn
    ctx.cls.info = type_info
    return ctx


class TestPluginFunction(test.TestCase):
    def test_returns_plugin_class(self):
        result = ovo_mypy.plugin('1.0')
        self.assertEqual(ovo_mypy.OsloVersionedObjectPlugin, result)


class TestGetClassDecoratorHook(test.TestCase):
    def setUp(self):
        super().setUp()
        self.plugin = _make_plugin()

    def test_returns_hook_for_versioned_object_registry(self):
        hook = self.plugin.get_class_decorator_hook(
            'oslo_versionedobjects.base.VersionedObjectRegistry.register'
        )
        self.assertIsNotNone(hook)
        self.assertTrue(callable(hook))

    def test_returns_none_for_non_matching(self):
        hook = self.plugin.get_class_decorator_hook('some.other.Decorator')
        self.assertIsNone(hook)

    def test_env_var_custom_decorator_matches(self):
        self.useFixture(
            fixtures.EnvironmentVariable(
                'OVO_MYPY_DECORATOR_CLASSES', 'MyCustomRegistry'
            )
        )
        hook = self.plugin.get_class_decorator_hook(
            'myproject.MyCustomRegistry.register'
        )
        self.assertIsNotNone(hook)

    def test_env_var_excludes_default_when_overridden(self):
        self.useFixture(
            fixtures.EnvironmentVariable(
                'OVO_MYPY_DECORATOR_CLASSES', 'MyCustomRegistry'
            )
        )
        hook = self.plugin.get_class_decorator_hook(
            'oslo_versionedobjects.base.VersionedObjectRegistry.register'
        )
        self.assertIsNone(hook)


class TestGetClassDecoratorHook2(test.TestCase):
    def setUp(self):
        super().setUp()
        self.plugin = _make_plugin()

    def test_returns_hook_for_versioned_object_registry(self):
        hook = self.plugin.get_class_decorator_hook_2(
            'oslo_versionedobjects.base.VersionedObjectRegistry.register'
        )
        self.assertIsNotNone(hook)
        self.assertTrue(callable(hook))

    def test_returns_none_for_non_matching(self):
        hook = self.plugin.get_class_decorator_hook_2('some.other.Decorator')
        self.assertIsNone(hook)

    def test_env_var_custom_decorator_matches(self):
        self.useFixture(
            fixtures.EnvironmentVariable(
                'OVO_MYPY_DECORATOR_CLASSES', 'MyCustomRegistry'
            )
        )
        hook = self.plugin.get_class_decorator_hook_2(
            'myproject.MyCustomRegistry.register'
        )
        self.assertIsNotNone(hook)

    def test_env_var_excludes_default_when_overridden(self):
        self.useFixture(
            fixtures.EnvironmentVariable(
                'OVO_MYPY_DECORATOR_CLASSES', 'MyCustomRegistry'
            )
        )
        hook = self.plugin.get_class_decorator_hook_2(
            'oslo_versionedobjects.base.VersionedObjectRegistry.register'
        )
        self.assertIsNone(hook)


class TestGetBaseClassHook(test.TestCase):
    def setUp(self):
        super().setUp()
        self.plugin = _make_plugin()

    def test_returns_hook_for_versioned_object(self):
        hook = self.plugin.get_base_class_hook(
            'oslo_versionedobjects.base.VersionedObject'
        )
        self.assertIsNotNone(hook)
        self.assertTrue(callable(hook))

    def test_returns_none_for_non_matching(self):
        hook = self.plugin.get_base_class_hook('some.other.BaseClass')
        self.assertIsNone(hook)

    def test_env_var_custom_base_class_matches(self):
        self.useFixture(
            fixtures.EnvironmentVariable(
                'OVO_MYPY_BASE_CLASSES', 'MyBaseObject'
            )
        )
        hook = self.plugin.get_base_class_hook('myproject.MyBaseObject')
        self.assertIsNotNone(hook)

    def test_env_var_excludes_default_when_overridden(self):
        self.useFixture(
            fixtures.EnvironmentVariable(
                'OVO_MYPY_BASE_CLASSES', 'MyBaseObject'
            )
        )
        hook = self.plugin.get_base_class_hook(
            'oslo_versionedobjects.base.VersionedObject'
        )
        self.assertIsNone(hook)

    def test_returns_none_for_builtins_object(self):
        hook = self.plugin.get_base_class_hook('builtins.object')
        self.assertIsNone(hook)


class TestAddMemberToClass(test.TestCase):
    def setUp(self):
        super().setUp()
        self.plugin = _make_plugin()

    def test_member_appears_in_symbol_table(self):
        type_info = _make_class_info('MyObj')
        member_type = types.AnyType(types.TypeOfAny.implementation_artifact)
        self.plugin._add_member_to_class('some_field', member_type, type_info)
        self.assertIn('some_field', type_info.names)

    def test_symbol_node_has_correct_type(self):
        type_info = _make_class_info('MyObj')
        member_type = types.AnyType(types.TypeOfAny.implementation_artifact)
        self.plugin._add_member_to_class('some_field', member_type, type_info)
        sym_node = type_info.names['some_field']
        self.assertIsInstance(sym_node.node, nodes.Var)
        self.assertEqual(member_type, sym_node.node.type)

    def test_symbol_node_has_correct_fullname(self):
        type_info = _make_class_info('MyObj')
        member_type = types.AnyType(types.TypeOfAny.implementation_artifact)
        self.plugin._add_member_to_class('some_field', member_type, type_info)
        sym_node = type_info.names['some_field']
        self.assertEqual('mymodule.MyObj.some_field', sym_node.node._fullname)


class TestGetPythonTypeFromOvoFieldType(test.TestCase):
    def setUp(self):
        super().setUp()
        self.plugin = _make_plugin()

    def test_returns_type_from_auto_typed_field_generic(self):
        expected = types.AnyType(types.TypeOfAny.special_form)
        ctx = _make_ctx()
        ctx.api.lookup_fully_qualified_or_none.return_value = _make_field_type(
            expected
        )
        result = self.plugin._get_python_type_from_ovo_field_type(
            ctx,
            'oslo_versionedobjects.fields.IntegerField',
            [],
            {},
        )
        self.assertIs(expected, result)

    def test_auto_typed_field_skips_typevar_arguments(self):
        any_type = types.AnyType(types.TypeOfAny.special_form)
        typevar = types.TypeVarType(
            'T', 'T', types.TypeVarId(-1), [], any_type, any_type
        )
        ctx = _make_ctx()
        ctx.api.lookup_fully_qualified_or_none.return_value = _make_field_type(
            typevar
        )
        result = self.plugin._get_python_type_from_ovo_field_type(
            ctx,
            'oslo_versionedobjects.fields.IntegerField',
            [],
            {},
        )
        self.assertIsInstance(result, types.AnyType)

    def test_returns_any_when_neither_generic_nor_mypy_type(self):
        type_info = _make_class_info(
            'CustomField', 'oslo_versionedobjects.fields'
        )
        type_info.mro = [type_info]
        field_symbol = nodes.SymbolTableNode(nodes.GDEF, type_info)
        ctx = mock.MagicMock()
        ctx.api.lookup_fully_qualified_or_none.return_value = field_symbol
        result = self.plugin._get_python_type_from_ovo_field_type(
            ctx,
            'oslo_versionedobjects.fields.CustomField',
            [],
            {},
        )
        self.assertIsInstance(result, types.AnyType)

    def test_returns_any_when_field_type_not_found(self):
        ctx = mock.MagicMock()
        ctx.api.lookup_fully_qualified_or_none.return_value = None
        result = self.plugin._get_python_type_from_ovo_field_type(
            ctx,
            'oslo_versionedobjects.fields.UnknownField',
            [],
            {},
        )
        self.assertIsInstance(result, types.AnyType)

    def test_returns_union_with_none_when_nullable_true(self):
        field_type = types.AnyType(types.TypeOfAny.special_form)
        ctx = _make_ctx()
        ctx.api.lookup_fully_qualified_or_none.return_value = _make_field_type(
            field_type
        )
        ctx.api.parse_bool.return_value = True
        result = self.plugin._get_python_type_from_ovo_field_type(
            ctx,
            'oslo_versionedobjects.fields.IntegerField',
            [],
            {'nullable': nodes.NameExpr('True')},
        )
        self.assertIsInstance(result, types.UnionType)
        self.assertTrue(
            any(isinstance(item, types.NoneType) for item in result.items)
        )

    def test_returns_plain_type_when_nullable_false(self):
        field_type = types.AnyType(types.TypeOfAny.special_form)
        ctx = _make_ctx()
        ctx.api.lookup_fully_qualified_or_none.return_value = _make_field_type(
            field_type
        )
        ctx.api.parse_bool.return_value = False
        result = self.plugin._get_python_type_from_ovo_field_type(
            ctx,
            'oslo_versionedobjects.fields.IntegerField',
            [],
            {'nullable': nodes.NameExpr('False')},
        )
        self.assertNotIsInstance(result, types.UnionType)


class TestAddOvoMembersToClass(test.TestCase):
    """Tests for _add_ovo_members_to_class, focusing on processed_fields."""

    def setUp(self):
        super().setUp()
        self.plugin = _make_plugin()

    def _make_ctx_with_any_api(self, name, statements):
        """Return a ClassDefContext whose API always resolves to AnyType."""
        ctx = _make_ctx(name, statements)
        ctx.api.lookup_fully_qualified_or_none.return_value = None
        return ctx

    def test_adds_fields_to_processed_set(self):
        assignment = _make_fields_assignment(
            ('my_id', 'oslo_versionedobjects.fields.IntegerField'),
            ('name', 'oslo_versionedobjects.fields.StringField'),
        )
        ctx = self._make_ctx_with_any_api('MyObj', [assignment])
        processed: set[str] = set()
        self.plugin._add_ovo_members_to_class(
            ctx, assignment.rvalue, processed
        )
        self.assertIn('my_id', processed)
        self.assertIn('name', processed)

    def test_skips_fields_already_in_processed_set(self):
        assignment = _make_fields_assignment(
            ('id', 'oslo_versionedobjects.fields.IntegerField'),
        )
        ctx = self._make_ctx_with_any_api('MyObj', [assignment])
        # Pre-populate processed_fields as if a derived class defined 'id'
        processed: set[str] = {'id'}
        self.plugin._add_ovo_members_to_class(
            ctx, assignment.rvalue, processed
        )
        # The field from the parent must not overwrite the derived class's
        self.assertNotIn('id', ctx.cls.info.names)

    def test_non_string_key_is_skipped_with_error(self):
        # Build a fields dict with a non-literal key: fields = {x: IntField()}
        key = nodes.NameExpr('x')
        callee = nodes.NameExpr('IntegerField')
        callee.fullname = 'oslo_versionedobjects.fields.IntegerField'
        call = nodes.CallExpr(callee, [], [], [])
        dict_expr = nodes.DictExpr([(key, call)])
        lvalue = nodes.NameExpr('fields')
        assignment = nodes.AssignmentStmt([lvalue], dict_expr)
        ctx = self._make_ctx_with_any_api('MyObj', [assignment])
        processed: set[str] = set()
        self.plugin._add_ovo_members_to_class(ctx, dict_expr, processed)
        ctx.api.fail.assert_called_once()

    def test_returns_true_when_all_fields_resolved(self):
        assignment = _make_fields_assignment(
            ('my_id', 'oslo_versionedobjects.fields.IntegerField'),
        )
        ctx = self._make_ctx_with_any_api('MyObj', [assignment])
        processed: set[str] = set()
        result = self.plugin._add_ovo_members_to_class(
            ctx, assignment.rvalue, processed
        )
        self.assertTrue(result)

    def test_returns_false_when_object_field_unresolvable(self):
        field_fullname = 'oslo_versionedobjects.fields.ObjectField'
        assignment = _make_object_field_assignment(
            'child', field_fullname, 'Missing'
        )
        ctx = _make_ctx('MyObj', [assignment])
        ctx.api.lookup_qualified.return_value = None
        ctx.api.modules = {}

        def _lookup_fqn(name):
            if name == field_fullname:
                cls_info = _make_class_info(
                    'ObjectField', 'oslo_versionedobjects.fields'
                )
                return nodes.SymbolTableNode(nodes.GDEF, cls_info)
            return None

        ctx.api.lookup_fully_qualified_or_none.side_effect = _lookup_fqn
        processed: set[str] = set()
        result = self.plugin._add_ovo_members_to_class(
            ctx, assignment.rvalue, processed
        )
        self.assertFalse(result)
        # Field is still added, typed as Any
        self.assertIn('child', ctx.cls.info.names)
        self.assertIsInstance(
            ctx.cls.info.names['child'].node.type, types.AnyType
        )

    def test_empty_fullname_treated_as_unresolved(self):
        # A callee with fullname='' (unresolved AST node) should be skipped
        # gracefully, yielding AnyType rather than a crash.
        key = nodes.StrExpr('my_field')
        callee = nodes.NameExpr('IntegerField')
        callee.fullname = ''
        call = nodes.CallExpr(callee, [], [], [])
        dict_expr = nodes.DictExpr([(key, call)])
        ctx = self._make_ctx_with_any_api('MyObj', [])
        processed: set[str] = set()
        self.plugin._add_ovo_members_to_class(ctx, dict_expr, processed)
        self.assertIn('my_field', ctx.cls.info.names)
        field_type = ctx.cls.info.names['my_field'].node.type
        self.assertIsInstance(field_type, types.AnyType)


class TestResolveOvoClassType(test.TestCase):
    def setUp(self):
        super().setUp()
        self.plugin = _make_plugin()

    def _make_ctx_with_qualified_lookup(self, type_info):
        ctx = mock.MagicMock()
        ctx.cls = mock.MagicMock()
        if type_info is not None:
            sym = nodes.SymbolTableNode(nodes.GDEF, type_info)
            ctx.api.lookup_qualified.return_value = sym
        else:
            ctx.api.lookup_qualified.return_value = None
        ctx.api.modules = {}
        return ctx

    def test_resolves_class_via_lookup_qualified(self):
        target_type_info = _make_class_info('HostVIFInfo', 'mymodule')
        ctx = self._make_ctx_with_qualified_lookup(target_type_info)
        result = self.plugin._resolve_ovo_class_type(ctx, 'HostVIFInfo')
        self.assertIsNotNone(result)
        self.assertIsInstance(result, types.Instance)
        self.assertIs(result.type, target_type_info)

    def test_falls_back_to_modules_when_lookup_qualified_fails(self):
        target_type_info = _make_class_info('HostVIFInfo', 'mymodule')
        ctx = mock.MagicMock()
        ctx.cls = mock.MagicMock()
        ctx.api.lookup_qualified.return_value = None
        sym = nodes.SymbolTableNode(nodes.GDEF, target_type_info)
        module = mock.MagicMock()
        module.names = {'HostVIFInfo': sym}
        ctx.api.modules = {'mymodule': module}
        result = self.plugin._resolve_ovo_class_type(ctx, 'HostVIFInfo')
        self.assertIsNotNone(result)
        self.assertIsInstance(result, types.Instance)

    def test_returns_none_when_not_found(self):
        ctx = mock.MagicMock()
        ctx.cls = mock.MagicMock()
        ctx.api.lookup_qualified.return_value = None
        ctx.api.modules = {}
        result = self.plugin._resolve_ovo_class_type(ctx, 'NotAClass')
        self.assertIsNone(result)

    def test_ignores_non_typeinfo_symbols(self):
        var = nodes.Var('HostVIFInfo')
        sym = nodes.SymbolTableNode(nodes.GDEF, var)
        ctx = mock.MagicMock()
        ctx.cls = mock.MagicMock()
        ctx.api.lookup_qualified.return_value = sym
        ctx.api.modules = {}
        result = self.plugin._resolve_ovo_class_type(ctx, 'HostVIFInfo')
        self.assertIsNone(result)


class TestGetPythonTypeObjectFields(test.TestCase):
    """Tests for ObjectField / ListOfObjectsField handling."""

    def setUp(self):
        super().setUp()
        self.plugin = _make_plugin()

    def _make_field_sym(self, field_fullname):
        """Return a SymbolTableNode for a field class."""
        cls_name = field_fullname.split('.')[-1]
        module = '.'.join(field_fullname.split('.')[:-1])
        type_info = _make_class_info(cls_name, module)
        return nodes.SymbolTableNode(nodes.GDEF, type_info)

    def _make_ctx_resolving(self, field_fullname, target_type_info):
        """Return a ctx resolving the field class and target object class."""
        field_sym = self._make_field_sym(field_fullname)
        target_sym = nodes.SymbolTableNode(nodes.GDEF, target_type_info)

        # Build a TypeInfo for builtins.list so Instance(list, [...]) works.
        list_sym_table = nodes.SymbolTable()
        list_block = nodes.Block([])
        list_cls_def = nodes.ClassDef('list', list_block)
        list_type_info = nodes.TypeInfo(
            list_sym_table, list_cls_def, 'builtins'
        )
        list_type_info._fullname = 'builtins.list'
        list_cls_def.info = list_type_info
        list_sym = nodes.SymbolTableNode(nodes.GDEF, list_type_info)

        ctx = mock.MagicMock()
        ctx.cls = mock.MagicMock()
        ctx.api.lookup_qualified.return_value = target_sym
        ctx.api.modules = {}
        ctx.api.parse_bool.return_value = False

        def _lookup_fqn(name):
            if name == field_fullname:
                return field_sym
            if name == 'builtins.list':
                return list_sym
            return None

        ctx.api.lookup_fully_qualified_or_none.side_effect = _lookup_fqn
        return ctx

    def _make_ctx_field_only(self, field_fullname):
        """Return a ctx where the field is found but target lookup fails."""
        field_sym = self._make_field_sym(field_fullname)

        ctx = mock.MagicMock()
        ctx.cls = mock.MagicMock()
        ctx.api.lookup_qualified.return_value = None
        ctx.api.modules = {}

        def _lookup_fqn(name):
            if name == field_fullname:
                return field_sym
            return None

        ctx.api.lookup_fully_qualified_or_none.side_effect = _lookup_fqn
        return ctx

    def test_object_field_returns_resolved_type(self):
        target = _make_class_info('HostVIFInfo', 'mymodule')
        ctx = self._make_ctx_resolving(
            'oslo_versionedobjects.fields.ObjectField', target
        )
        result = self.plugin._get_python_type_from_ovo_field_type(
            ctx,
            'oslo_versionedobjects.fields.ObjectField',
            [nodes.StrExpr('HostVIFInfo')],
            {},
        )
        self.assertIsInstance(result, types.Instance)
        self.assertIs(result.type, target)

    def test_list_of_objects_field_returns_list_of_resolved_type(self):
        target = _make_class_info('HostVIFInfo', 'mymodule')
        ctx = self._make_ctx_resolving(
            'oslo_versionedobjects.fields.ListOfObjectsField', target
        )
        result = self.plugin._get_python_type_from_ovo_field_type(
            ctx,
            'oslo_versionedobjects.fields.ListOfObjectsField',
            [nodes.StrExpr('HostVIFInfo')],
            {},
        )
        self.assertIsInstance(result, types.Instance)
        self.assertEqual('builtins.list', result.type.fullname)
        self.assertEqual(1, len(result.args))
        self.assertIsInstance(result.args[0], types.Instance)
        self.assertIs(result.args[0].type, target)

    def test_object_field_nullable_returns_union_with_none(self):
        target = _make_class_info('HostVIFInfo', 'mymodule')
        ctx = self._make_ctx_resolving(
            'oslo_versionedobjects.fields.ObjectField', target
        )
        ctx.api.parse_bool.return_value = True
        result = self.plugin._get_python_type_from_ovo_field_type(
            ctx,
            'oslo_versionedobjects.fields.ObjectField',
            [nodes.StrExpr('HostVIFInfo')],
            {'nullable': nodes.NameExpr('True')},
        )
        self.assertIsInstance(result, types.UnionType)
        self.assertTrue(
            any(isinstance(t, types.NoneType) for t in result.items)
        )

    def test_list_of_objects_field_nullable_returns_union_with_none(self):
        target = _make_class_info('HostVIFInfo', 'mymodule')
        ctx = self._make_ctx_resolving(
            'oslo_versionedobjects.fields.ListOfObjectsField', target
        )
        ctx.api.parse_bool.return_value = True
        result = self.plugin._get_python_type_from_ovo_field_type(
            ctx,
            'oslo_versionedobjects.fields.ListOfObjectsField',
            [nodes.StrExpr('HostVIFInfo')],
            {'nullable': nodes.NameExpr('True')},
        )
        self.assertIsInstance(result, types.UnionType)
        self.assertTrue(
            any(isinstance(t, types.NoneType) for t in result.items)
        )
        # The non-None item should be list[HostVIFInfo], not list[X | None]
        non_none = [
            t for t in result.items if not isinstance(t, types.NoneType)
        ]
        self.assertEqual(1, len(non_none))
        self.assertIsInstance(non_none[0], types.Instance)
        self.assertEqual('builtins.list', non_none[0].type.fullname)

    def test_object_field_unresolvable_class_returns_none(self):
        ctx = self._make_ctx_field_only(
            'oslo_versionedobjects.fields.ObjectField'
        )
        result = self.plugin._get_python_type_from_ovo_field_type(
            ctx,
            'oslo_versionedobjects.fields.ObjectField',
            [nodes.StrExpr('NoSuchClass')],
            {},
        )
        self.assertIsNone(result)

    def test_object_field_non_string_arg_returns_none(self):
        ctx = self._make_ctx_field_only(
            'oslo_versionedobjects.fields.ObjectField'
        )
        result = self.plugin._get_python_type_from_ovo_field_type(
            ctx,
            'oslo_versionedobjects.fields.ObjectField',
            [nodes.NameExpr('some_var')],
            {},
        )
        self.assertIsNone(result)

    def test_object_field_no_positional_args_returns_none(self):
        ctx = self._make_ctx_field_only(
            'oslo_versionedobjects.fields.ObjectField'
        )
        result = self.plugin._get_python_type_from_ovo_field_type(
            ctx,
            'oslo_versionedobjects.fields.ObjectField',
            [],
            {},
        )
        self.assertIsNone(result)


class TestAddOvoMembersObjectField(test.TestCase):
    """Integration tests: _add_ovo_members_to_class with ObjectField."""

    def setUp(self):
        super().setUp()
        self.plugin = _make_plugin()

    def _make_ctx_with_object_resolution(
        self, field_fullname, target_type_info
    ):
        cls_name = field_fullname.split('.')[-1]
        module = '.'.join(field_fullname.split('.')[:-1])
        field_type_info = _make_class_info(cls_name, module)
        field_sym = nodes.SymbolTableNode(nodes.GDEF, field_type_info)

        target_sym = nodes.SymbolTableNode(nodes.GDEF, target_type_info)
        list_sym_table = nodes.SymbolTable()
        list_block = nodes.Block([])
        list_cls_def = nodes.ClassDef('list', list_block)
        list_type_info = nodes.TypeInfo(
            list_sym_table, list_cls_def, 'builtins'
        )
        list_type_info._fullname = 'builtins.list'
        list_cls_def.info = list_type_info
        list_sym = nodes.SymbolTableNode(nodes.GDEF, list_type_info)

        ctx = _make_ctx('Owner', [])
        ctx.api.lookup_qualified.return_value = target_sym
        ctx.api.modules = {}
        ctx.api.parse_bool.return_value = False

        def _lookup_fqn(name):
            if name == field_fullname:
                return field_sym
            if name == 'builtins.list':
                return list_sym
            return None

        ctx.api.lookup_fully_qualified_or_none.side_effect = _lookup_fqn
        return ctx

    def test_list_of_objects_field_resolved_to_list_type(self):
        field_fullname = 'oslo_versionedobjects.fields.ListOfObjectsField'
        target = _make_class_info('ChildObj', 'mymodule')
        assignment = _make_object_field_assignment(
            'children',
            field_fullname,
            'ChildObj',
        )
        ctx = self._make_ctx_with_object_resolution(field_fullname, target)
        processed: set[str] = set()
        self.plugin._add_ovo_members_to_class(
            ctx, assignment.rvalue, processed
        )
        self.assertIn('children', ctx.cls.info.names)
        field_type = ctx.cls.info.names['children'].node.type
        self.assertIsInstance(field_type, types.Instance)
        self.assertEqual('builtins.list', field_type.type.fullname)

    def test_object_field_resolved_to_instance_type(self):
        field_fullname = 'oslo_versionedobjects.fields.ObjectField'
        target = _make_class_info('ChildObj', 'mymodule')
        assignment = _make_object_field_assignment(
            'child',
            field_fullname,
            'ChildObj',
        )
        ctx = self._make_ctx_with_object_resolution(field_fullname, target)
        processed: set[str] = set()
        self.plugin._add_ovo_members_to_class(
            ctx, assignment.rvalue, processed
        )
        self.assertIn('child', ctx.cls.info.names)
        field_type = ctx.cls.info.names['child'].node.type
        self.assertIsInstance(field_type, types.Instance)
        self.assertIs(field_type.type, target)


class TestGenerateOvoFieldDefs(test.TestCase):
    def setUp(self):
        super().setUp()
        self.plugin = _make_plugin()

    def test_no_fields_dict_is_noop(self):
        ctx = _make_ctx('MyObj', [])
        ctx.cls.info.mro = [ctx.cls.info]
        self.plugin.generate_ovo_field_defs(ctx)
        self.assertEqual({}, dict(ctx.cls.info.names))

    def test_fields_are_added_to_class(self):
        field_type = types.AnyType(types.TypeOfAny.special_form)
        ctx = _make_ctx(
            'MyObj',
            [
                _make_fields_assignment(
                    ('my_id', 'oslo_versionedobjects.fields.IntegerField'),
                    ('name', 'oslo_versionedobjects.fields.StringField'),
                )
            ],
        )
        ctx.cls.info.mro = [ctx.cls.info]
        ctx.api.lookup_fully_qualified_or_none.return_value = _make_field_type(
            field_type
        )
        ctx.api.parse_bool.return_value = False
        self.plugin.generate_ovo_field_defs(ctx)
        self.assertIn('my_id', ctx.cls.info.names)
        self.assertIn('name', ctx.cls.info.names)

    def test_unknown_field_type_defaults_to_any(self):
        ctx = _make_ctx(
            'MyObj',
            [
                _make_fields_assignment(
                    ('magic', 'myproject.fields.MagicField'),
                )
            ],
        )
        ctx.cls.info.mro = [ctx.cls.info]
        ctx.api.lookup_fully_qualified_or_none.return_value = None
        self.plugin.generate_ovo_field_defs(ctx)
        self.assertIn('magic', ctx.cls.info.names)
        sym_node = ctx.cls.info.names['magic']
        self.assertIsInstance(sym_node.node.type, types.AnyType)

    def test_inherited_fields_are_included(self):
        """Fields from a parent class in the MRO are added to the child."""
        field_type = types.AnyType(types.TypeOfAny.special_form)
        parent_type_info = _make_class_info('Base', 'mymodule')
        parent_type_info.defn.defs.body = [
            _make_fields_assignment(
                ('inherited_id', 'oslo_versionedobjects.fields.IntegerField'),
            )
        ]
        ctx = _make_ctx(
            'MyObj',
            [
                _make_fields_assignment(
                    ('name', 'oslo_versionedobjects.fields.StringField'),
                )
            ],
        )
        ctx.cls.info.mro = [ctx.cls.info, parent_type_info]
        ctx.api.lookup_fully_qualified_or_none.return_value = _make_field_type(
            field_type
        )
        ctx.api.parse_bool.return_value = False
        self.plugin.generate_ovo_field_defs(ctx)
        self.assertIn('name', ctx.cls.info.names)
        self.assertIn('inherited_id', ctx.cls.info.names)

    def test_child_field_wins_over_same_named_parent_field(self):
        """When child and parent both define a field, the child's wins."""
        field_type = types.AnyType(types.TypeOfAny.special_form)
        parent_type_info = _make_class_info('Base', 'mymodule')
        parent_type_info.defn.defs.body = [
            _make_fields_assignment(
                ('shared', 'oslo_versionedobjects.fields.StringField'),
            )
        ]
        ctx = _make_ctx(
            'MyObj',
            [
                _make_fields_assignment(
                    ('shared', 'oslo_versionedobjects.fields.IntegerField'),
                )
            ],
        )
        ctx.cls.info.mro = [ctx.cls.info, parent_type_info]
        ctx.api.lookup_fully_qualified_or_none.return_value = _make_field_type(
            field_type
        )
        ctx.api.parse_bool.return_value = False
        self.plugin.generate_ovo_field_defs(ctx)
        # The field must appear exactly once in the child class's names
        self.assertIn('shared', ctx.cls.info.names)

    def test_mixin_fields_included_via_body(self):
        """Fields from a non-OVO mixin are picked up from its class body."""
        field_type = types.AnyType(types.TypeOfAny.special_form)
        mixin_type_info = _make_class_info('TimestampMixin', 'mymodule')
        mixin_type_info.defn.defs.body = [
            _make_fields_assignment(
                ('created_at', 'oslo_versionedobjects.fields.DateTimeField'),
                ('updated_at', 'oslo_versionedobjects.fields.DateTimeField'),
            )
        ]
        ctx = _make_ctx(
            'MyObj',
            [
                _make_fields_assignment(
                    ('name', 'oslo_versionedobjects.fields.StringField'),
                )
            ],
        )
        ctx.cls.info.mro = [ctx.cls.info, mixin_type_info]
        ctx.api.lookup_fully_qualified_or_none.return_value = _make_field_type(
            field_type
        )
        ctx.api.parse_bool.return_value = False
        self.plugin.generate_ovo_field_defs(ctx)
        self.assertIn('name', ctx.cls.info.names)
        self.assertIn('created_at', ctx.cls.info.names)
        self.assertIn('updated_at', ctx.cls.info.names)


class TestGenerateOvoFieldDefs2(test.TestCase):
    def setUp(self):
        super().setUp()
        self.plugin = _make_plugin()

    def test_returns_true_when_all_fields_resolved(self):
        field_type = types.AnyType(types.TypeOfAny.special_form)
        ctx = _make_ctx(
            'MyObj',
            [
                _make_fields_assignment(
                    ('my_id', 'oslo_versionedobjects.fields.IntegerField'),
                )
            ],
        )
        ctx.cls.info.mro = [ctx.cls.info]
        ctx.api.lookup_fully_qualified_or_none.return_value = _make_field_type(
            field_type
        )
        ctx.api.parse_bool.return_value = False
        result = self.plugin.generate_ovo_field_defs_2(ctx)
        self.assertTrue(result)

    def test_returns_true_even_when_object_field_unresolvable(self):
        # hook_2 always returns True: by the time it fires all modules are
        # loaded, so retrying (False) would not help an unresolvable target.
        field_fullname = 'oslo_versionedobjects.fields.ObjectField'
        ctx = _make_ctx(
            'MyObj',
            [
                _make_object_field_assignment(
                    'child', field_fullname, 'Missing'
                )
            ],
        )
        ctx.cls.info.mro = [ctx.cls.info]
        ctx.api.lookup_qualified.return_value = None
        ctx.api.modules = {}

        def _lookup_fqn(name):
            if name == field_fullname:
                cls_info = _make_class_info(
                    'ObjectField', 'oslo_versionedobjects.fields'
                )
                return nodes.SymbolTableNode(nodes.GDEF, cls_info)
            return None

        ctx.api.lookup_fully_qualified_or_none.side_effect = _lookup_fqn
        result = self.plugin.generate_ovo_field_defs_2(ctx)
        self.assertTrue(result)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_versionedobjects import base
from oslo_versionedobjects import fields


@base.VersionedObjectRegistry.register
class CompoundFieldsObject(base.VersionedObject):
    VERSION = '1.0'
    fields = {
        # AutoTypedField subclasses — plugin infers exact container types
        'str_list': fields.ListOfStringsField(),
        'str_set': fields.SetOfStringsField(),
        'str_dict': fields.DictOfStringsField(),
        'int_dict': fields.DictOfIntegersField(),
        'nested_list': fields.ListOfListsOfStringsField(),
        'states': fields.ListOfEnumField(['active', 'inactive']),
        'nullable_list': fields.ListOfStringsField(nullable=True),
        'nullable_str_dict': fields.DictOfNullableStringsField(),
        # Raw compound field types (Field wrapping a CompoundFieldType) — the
        # plugin only handles AutoTypedField subclasses, so these fall back to
        # Any.  Use a dedicated AutoTypedField subclass to get a precise type.
        'raw_list': fields.Field(fields.List(fields.Integer())),
        'raw_dict': fields.Field(fields.Dict(fields.String())),
        'raw_set': fields.Field(fields.Set(fields.Integer())),
    }


obj = CompoundFieldsObject()
reveal_type(obj.str_list)  # N: Revealed type is "list[str]"
reveal_type(obj.str_set)  # N: Revealed type is "set[str]"
reveal_type(obj.str_dict)  # N: Revealed type is "dict[str, str]"
reveal_type(obj.int_dict)  # N: Revealed type is "dict[str, int]"
reveal_type(obj.nested_list)  # N: Revealed type is "list[list[str]]"
reveal_type(obj.states)  # N: Revealed type is "list[str]"
reveal_type(obj.nullable_list)  # N: Revealed type is "list[str] | None"
reveal_type(
    obj.nullable_str_dict  # N: Revealed type is "dict[str, str | None]"
)
reveal_type(obj.raw_list)  # N: Revealed type is "Any"
reveal_type(obj.raw_dict)  # N: Revealed type is "Any"
reveal_type(obj.raw_set)  # N: Revealed type is "Any"



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_versionedobjects import base
from oslo_versionedobjects import fields


class StatusField(fields.BaseEnumField):
    """A named enum with a fixed set of values."""

    AUTO_TYPE = fields.Enum(valid_values=['active', 'error', 'deleted'])


@base.VersionedObjectRegistry.register
class EnumFieldsObject(base.VersionedObject):
    VERSION = '1.0'
    fields = {
        # Anonymous EnumField and its named BaseEnumField subclass both resolve
        # to Any — the generic parameter on BaseEnumField is explicitly Any
        # because enum values are strings at runtime but the field accepts any
        # valid value type.
        'status': StatusField(),
        'state': fields.EnumField(['pending', 'running', 'done']),
        'nullable_status': StatusField(nullable=True),
    }


obj = EnumFieldsObject()
reveal_type(obj.status)  # N: Revealed type is "Any"
reveal_type(obj.state)  # N: Revealed type is "Any"
reveal_type(obj.nullable_status)  # N: Revealed type is "Any | None"



# Copyright 2010 United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Fixtures for VersionedObject tests."""

import gettext
import logging
import os
from typing import Any, cast
import warnings

import fixtures
from oslo_config import cfg


_TRUE_VALUES = ('True', 'true', '1', 'yes')

CONF = cfg.CONF
DB_SCHEMA = ""


class TranslationFixture(fixtures.Fixture):
    """Use gettext NullTranslation objects in tests."""

    def setUp(self) -> None:
        super().setUp()
        nulltrans = gettext.NullTranslations()
        gettext_fixture = fixtures.MonkeyPatch(
            'gettext.translation', lambda *x, **y: nulltrans
        )
        self.gettext_patcher = self.useFixture(gettext_fixture)


class NullHandler(logging.Handler):
    """custom default NullHandler to attempt to format the record.

    Used in conjunction with
    log_fixture.get_logging_handle_error_fixture to detect formatting errors in
    debug level logs without saving the logs.
    """

    def handle(self, record):
        self.format(record)

    def emit(self, record):
        pass

    def createLock(self) -> None:
        self.lock = None


class StandardLogging(fixtures.Fixture):
    """Setup Logging redirection for tests.

    There are a number of things we want to handle with logging in tests:

    * Redirect the logging to somewhere that we can test or dump it later.

    * Ensure that as many DEBUG messages as possible are actually
       executed, to ensure they are actually syntactically valid (they
       often have not been).

    * Ensure that we create useful output for tests that doesn't
      overwhelm the testing system (which means we can't capture the
      100 MB of debug logging on every run).

    To do this we create a logger fixture at the root level, which
    defaults to INFO and create a Null Logger at DEBUG which lets
    us execute log messages at DEBUG but not keep the output.

    To support local debugging OS_DEBUG=True can be set in the
    environment, which will print out the full debug logging.

    There are also a set of overrides for particularly verbose
    modules to be even less than INFO.

    """

    def setUp(self) -> None:
        super().setUp()

        # set root logger to debug
        root = logging.getLogger()
        root.setLevel(logging.DEBUG)

        # supports collecting debug level for local runs
        if os.environ.get('OS_DEBUG') in _TRUE_VALUES:
            level = logging.DEBUG
        else:
            level = logging.INFO

        # Collect logs
        fs = '%(asctime)s %(levelname)s [%(name)s] %(message)s'
        self.logger = self.useFixture(
            fixtures.FakeLogger(format=fs, level=None)
        )
        # TODO(sdague): why can't we send level through the fake
        # logger? Tests prove that it breaks, but it's worth getting
        # to the bottom of.
        root.handlers[0].setLevel(level)

        if level > logging.DEBUG:
            # Just attempt to format debug level logs, but don't save them
            handler = NullHandler()
            self.useFixture(fixtures.LogHandler(handler, nuke_handlers=False))
            handler.setLevel(logging.DEBUG)


class OutputStreamCapture(fixtures.Fixture):
    """Capture output streams during tests.

    This fixture captures errant printing to stderr / stdout during
    the tests and lets us see those streams at the end of the test
    runs instead. Useful to see what was happening during failed
    tests.
    """

    def setUp(self) -> None:
        super().setUp()
        if os.environ.get('OS_STDOUT_CAPTURE') in _TRUE_VALUES:
            self.out = self.useFixture(fixtures.StringStream('stdout'))
            self.useFixture(
                fixtures.MonkeyPatch('sys.stdout', self.out.stream)
            )
        if os.environ.get('OS_STDERR_CAPTURE') in _TRUE_VALUES:
            self.err = self.useFixture(fixtures.StringStream('stderr'))
            self.useFixture(
                fixtures.MonkeyPatch('sys.stderr', self.err.stream)
            )

    @property
    def stderr(self) -> str | None:
        if (
            self.err._details is None
            or self.err._details.get('stderr') is None
        ):
            return None
        return cast(str, self.err._details["stderr"].as_text())

    @property
    def stdout(self) -> str | None:
        if (
            self.out._details is None
            or self.out._details.get('stdout') is None
        ):
            return None
        return cast(str, self.out._details['stdout'].as_text())


class Timeout(fixtures.Fixture):
    """Setup per test timeouts.

    In order to avoid test deadlocks we support setting up a test
    timeout parameter read from the environment. In almost all
    cases where the timeout is reached this means a deadlock.

    A class level TIMEOUT_SCALING_FACTOR also exists, which allows
    extremely long tests to specify they need more time.
    """

    def __init__(self, timeout: Any, scaling: int = 1) -> None:
        super().__init__()
        try:
            self.test_timeout = int(timeout)
        except ValueError:
            # If timeout value is invalid do not set a timeout.
            self.test_timeout = 0
        if scaling >= 1:
            self.test_timeout *= scaling
        else:
            raise ValueError('scaling value must be >= 1')

    def setUp(self) -> None:
        super().setUp()
        if self.test_timeout > 0:
            self.useFixture(fixtures.Timeout(self.test_timeout, gentle=True))


class WarningsFixture(fixtures.Fixture):
    """Filters out warnings during test runs."""

    def setUp(self) -> None:
        super().setUp()
        # NOTE(sdague): Make deprecation warnings only happen once. Otherwise
        # this gets kind of crazy given the way that upstream python libs use
        # this.
        warnings.simplefilter("once", DeprecationWarning)

        self.addCleanup(warnings.resetwarnings)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Integration tests for the mypy plugin.

These tests run mypy programmatically against fixture files in the
``test_data/`` directory and verify that field types are inferred correctly
by the plugin.

Expected types are declared as inline ``# N: Revealed type is "..."``
annotations alongside each ``reveal_type()`` call in the fixture files,
mirroring the convention used by mypy's own test harness. The test
infrastructure parses those annotations and compares them against the
structured JSON output from mypy, avoiding fragile stdout regex parsing.
"""

import json
import pathlib
import re
import tempfile
from typing import TypedDict

import mypy.api
from oslo_versionedobjects import test


class _MypyMessage(TypedDict):
    file: str
    line: int
    column: int
    end_line: int
    end_column: int
    message: str
    hint: str | None
    code: str | None
    severity: str


# Absolute path to the project root (contains pyproject.toml with the plugin
# configuration).
_PROJECT_ROOT = pathlib.Path(__file__).parent.parent.parent.resolve()
_TEST_DATA_DIR = pathlib.Path(__file__).parent / 'test_data'


def _run_mypy(fixture_name: str) -> tuple[list[_MypyMessage], int]:
    """Run mypy on a fixture file and return (messages, exit_code).

    Messages is the full list of structured JSON objects emitted by mypy
    (one per stdout line).  Each dict has at least ``severity``, ``line``,
    and ``message`` keys.
    """
    fixture_path = _TEST_DATA_DIR / fixture_name
    with tempfile.TemporaryDirectory() as cache_dir:
        stdout, _stderr, exit_code = mypy.api.run(
            [
                '--no-incremental',
                '--cache-dir',
                cache_dir,
                '--output',
                'json',
                '--config-file',
                str(_PROJECT_ROOT / 'pyproject.toml'),
                str(fixture_path),
            ]
        )
    messages: list[_MypyMessage] = [
        json.loads(line) for line in stdout.splitlines() if line.strip()
    ]
    return messages, exit_code


def _get_expected_notes(fixture_name: str) -> dict[int, str]:
    """Return ``{line_number: expected_message}`` from ``# N:`` annotations.

    Scans the fixture file for inline comments of the form::

        reveal_type(obj.x)  # N: Revealed type is "str"

    and returns a mapping from the 1-based line number to the expected note
    message string.
    """
    fixture_path = _TEST_DATA_DIR / fixture_name
    notes: dict[int, str] = {}
    for lineno, line in enumerate(
        fixture_path.read_text().splitlines(), start=1
    ):
        m = re.search(r'#\s*N:\s*(.+)', line)
        if m:
            notes[lineno] = m.group(1).strip()
    return notes


class TestMypyIntegration(test.TestCase):
    def assertNoErrors(self, messages: list[_MypyMessage]) -> None:
        errors = [m for m in messages if m['severity'] == 'error']
        self.assertEqual([], errors, f"Unexpected mypy errors: {errors}")

    def assertNotes(self, fixture_name: str) -> None:
        """Assert that mypy emits the notes declared in the fixture file."""
        messages, _exit_code = _run_mypy(fixture_name)
        self.assertNoErrors(messages)

        actual = {
            m['line']: m['message']
            for m in messages
            if m['severity'] == 'note'
            and m['line'] > 0
            and m['message'].startswith('Revealed type is')
        }

        for lineno, expected_msg in _get_expected_notes(fixture_name).items():
            self.assertIn(
                lineno,
                actual,
                f"No reveal_type note at line {lineno} of {fixture_name}",
            )
            self.assertEqual(
                expected_msg,
                actual[lineno],
                f"Wrong revealed type at line {lineno} of {fixture_name}",
            )

    def test_empty_fields(self):
        """A model with no fields should be accepted without errors."""
        messages, _exit_code = _run_mypy('empty_fields.py')
        self.assertNoErrors(messages)

    def test_basic_autotyped_fields(self):
        """Standard AutoTypedField subclasses should map to the right types."""
        self.assertNotes('basic_fields.py')

    def test_subclassing_inherits_parent_fields(self):
        """A child class should expose fields defined on the parent class."""
        self.assertNotes('subclassing.py')

    def test_timestamp_mixin_fields(self):
        """Fields from base.TimestampedObject are inherited correctly."""
        self.assertNotes('timestamp_mixin.py')

    def test_compound_fields(self):
        """AutoTypedField compound types infer exact container element types.

        Raw Field(List/Dict/Set(...)) wrappers fall back to Any because the
        plugin only resolves types through the AutoTypedField generic
        parameter.
        """
        self.assertNotes('compound_fields.py')

    def test_enum_fields(self):
        """Enum fields resolve to Any because BaseEnumField is typed Any."""
        self.assertNotes('enum_fields.py')

    def test_object_fields(self):
        """ObjectField resolves to the concrete named class, not the base."""
        self.assertNotes('object_fields.py')



# Copyright 2010 United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# Copyright 2011 Justin Santa Barbara
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Utilities and helper functions."""

import datetime

# ISO 8601 extended time format without microseconds
_ISO8601_TIME_FORMAT: str = '%Y-%m-%dT%H:%M:%S'


def isotime(at: datetime.datetime) -> str:
    """Stringify time in ISO 8601 format."""
    st = at.strftime(_ISO8601_TIME_FORMAT)
    tz: str | None = at.tzinfo.tzname(None) if at.tzinfo else 'UTC'
    # Need to handle either iso8601 or python UTC format
    if tz is None or tz in ['UTC', 'UTC+00:00']:
        st += 'Z'
    else:
        st += tz
    return st
